import numpy as np

from coolbox.utilities import GenomeRange
from .base import HicMatBase


class Cool(HicMatBase):
    """
    Cool track from .cool, .mcool file.

    Parameters
    ----------
    file: str
        The file path of .cool, .mcool file

    balance: bool
        If use the balanced contact matrix.


    """

    DEFAULT_PROPERTIES = {
        'cmap': "JuiceBoxLike",
        'balance': True,
        "norm": "no",
        "transform": "log"
    }

    def __init__(self, file, **kwargs):
        properties = Cool.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs
        })
        super().__init__(**properties)

    def fetch_data(self, gr: GenomeRange, **kwargs) -> np.ndarray:
        from coolbox.utilities.hic.wrap import CoolerWrap

        path = self.properties['file']
        wrap = CoolerWrap(path, balance=self.balance, binsize=kwargs.get('resolution', 'auto'))
        arr = wrap.fetch(gr, kwargs.get('gr2'))

        self.fetched_binsize = wrap.fetched_binsize  # expose fetched binsize

        return self.fill_zero_nan(arr)

    def fetch_pixels(self, gr: GenomeRange, **kwargs):
        """
        Fetch the pixels table of upper triangle of the original contact matrix(not processed).

        Parameters
        ----------
        gr2 : GenomeRange, optional.

        balance : bool, optional
            balance matrix or not,
            default `self.is_balance`.

        resolution : {'auto', int}
            resolution of the data. for example 5000.
            'auto' for calculate resolution automatically.
            default 'auto'

        join : bool
            whether to expand the bin ID columns
            into (chrom, start, end).
            default True

        Returns
        -------
        pixels : pandas.core.frame.DataFrame
            Hi-C pixels table.
            The pixel table contains the non-zero upper triangle entries of the contact map.
        """
        from coolbox.utilities.hic.wrap import CoolerWrap

        path = self.properties['file']
        balance = kwargs.get('balance', self.is_balance)
        wrap = CoolerWrap(path, balance=balance, binsize=kwargs.get('resolution', 'auto'))

        pixels = wrap.fetch_pixels(gr, kwargs.get('gr2'), join=kwargs.get('join', True))
        return pixels

    def infer_binsize(self, gr: GenomeRange, **kwargs) -> int:
        from coolbox.utilities.hic.wrap import CoolerWrap

        path = self.properties['file']
        wrap = CoolerWrap(path, balance=self.balance, binsize=kwargs.get('resolution', 'auto'))
        return wrap.infer_binsize(gr)



import pandas as pd

from .base import ArcsBase
from .fetch import FetchParix
from coolbox.utilities.bed import process_bedpe
from coolbox.utilities.genome import GenomeRange


class BEDPE(ArcsBase, FetchParix):
    """
    Arcs track from .bedpe file.

    Parameters
    ----------
    file: str
        Path of .bedpe file

    pos : str, optional
        Method for choosing arch anchor for bedpe data: 'start', 'end', 'mid', default 'mid'

    """
    DEFAULT_PROPERTIES = {
        'pos': 'mid'
    }
    FIELDS = ["chrom1", "start1", "end1", "chrom2", "start2", "end2",
              "name", "score", "strand1", "strand2"]

    def __init__(self, file, **kwargs):
        properties = BEDPE.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs
        })
        super().__init__(**properties)
        self.bgz_file = process_bedpe(file)

    def fetch_data(self, gr: GenomeRange, **kwargs) -> pd.DataFrame:
        # filter peaks manually for hicpeaks style in fetch_plot_data
        df = self.fetch_intervals(self.bgz_file, gr, kwargs.get('gr2'))
        # TODO the returned df has no named columns, may cause error
        if len(df) == 0:
            return df

        columns = list(df.columns)
        for i, col in enumerate(self.FIELDS):
            if i >= len(columns):
                break
            columns[i] = col
        df.columns = columns
        for col in ['start1', 'end1', 'start2', 'end2']:
            df[col] = df[col].astype(int)
        if 'score' in df.columns:
            df['score'] = df['score'].astype(float)
        return df

    def fetch_plot_data(self, gr: GenomeRange, **kwargs) -> pd.DataFrame:
        df = self.fetch_data(gr, **kwargs)
        if len(df) == 0:
            return df

        style = self.properties['style']
        if style == self.STYLE_ARCS:
            pos_at = self.properties['pos']
            if pos_at == 'start':
                pos1, pos2 = df['start1'], df['start2']
            elif pos_at == 'end':
                pos1, pos2 = df['end1'], df['end2']
            else:
                pos1, pos2 = (df['end1'] + df['start1']) / 2, (df['end2'] + df['start2']) / 2
            return pd.DataFrame({'pos1': pos1, 'pos2': pos2, 'score': df['score']})
        elif style == self.STYLE_HICPEAKS:
            gr2 = kwargs.get('gr2')
            if gr2 and gr2 != gr:
                mask = (df['start2'] >= gr2.start) & (df['end2'] <= gr2.end)
                df = df[mask]
            return df
        else:
            raise ValueError("The supported style for bedpe data are ['arcs', 'hicpeaks']")



import random
import re

import pandas as pd
from dna_features_viewer import GraphicFeature, GraphicRecord

from coolbox.utilities import (
    split_genome_range, get_logger, GenomeRange
)
from coolbox.utilities.bed import (
    build_gtf_index, tabix_query
)
from .base import Track

log = get_logger(__name__)


class GTF(Track):
    """
    GTF gene annotation track.

    Parameters
    ----------
    file : str
        Path to .gtf(or .gtf.bgz) file.

    row_filter : str, optional
        Row filter expression, only keep the rows for draw. (Default 'feature == "gene"')

    length_ratio_thresh : float
        Length ratio threshold of features, (Default 0.01)

    color : {str, 'random'}, optional
        When the color is random, color for each gene will be randomly selected.
    """
    RANDOM_COLORS = [
        "#ffcccc",
        "#ffd700",
        "#cffccc",
        "#ff9c9c",
        "#66ccff"
    ]

    DEFAULT_PROPERTIES = {
        "height": 4,
        "color": "random",
        "row_filter": 'feature == "gene"',
        "length_ratio_thresh": 0.005,
    }

    def __init__(self, file, **kwargs):
        properties = GTF.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs
        })
        super().__init__(properties)
        self.bgz_file = build_gtf_index(file)
        color = self.properties['color']
        if (type(color) is str) and (color.startswith('#')):
            self.colors = [color]
        elif type(color) is list:
            self.colors = [c for c in color if (type(c) is str) and c.startswith('#')]
            if len(self.colors) == 0:
                self.colors = GTF.RANDOM_COLORS
        else:
            self.colors = GTF.RANDOM_COLORS

    def fetch_data(self, gr: GenomeRange, **kwargs):
        """
        Returns
        -------
        df: pandas.DataFrame
            should be with the format like:

            columns = ['seqname', 'source', 'feature', 'start', 'end',
                        'score', 'strand', 'frame', 'attribute', 'gene_name']

        """
        return self.fetch_intervals(gr)

    def fetch_intervals(self, gr: GenomeRange):
        """

        Parameters
        ----------
        gr : {str, GenomeRange}

        Returns
        -------
        intervals : pandas.core.frame.DataFrame
            Annotation interval table.
        """
        rows = []
        for row in tabix_query(self.bgz_file, gr.chrom, gr.start, gr.end):
            rows.append(row)
        if len(rows) == 0:
            gr.change_chrom_names()
            for row in tabix_query(self.bgz_file, gr.chrom, gr.start, gr.end):
                rows.append(row)

        columns = ['seqname', 'source', 'feature', 'start', 'end', 'score', 'strand', 'frame', 'attribute']
        df = pd.DataFrame(rows, columns=columns)
        df['start'] = df['start'].astype(int)
        df['end'] = df['end'].astype(int)
        df['gene_name'] = df['attribute'].str.extract(".*gene_name (.*?) ").iloc[:, 0].str.strip('\";')
        df['gene_name'].fillna("", inplace=True)
        return df

    def plot(self, ax, gr: GenomeRange, **kwargs):
        self.ax = ax
        df = self.fetch_plot_data(gr)
        if self.has_prop("row_filter"):
            filters = self.properties["row_filter"]
            for filter_ in filters.split(";"):
                try:
                    op_idx = list(re.finditer("[=><!]", filter_))[0].start()
                    l_ = filter_[:op_idx].strip()
                    r_ = filter_[op_idx:]
                    df = eval(f'df[df["{l_}"]{r_}]')
                except IndexError:
                    log.warning(f"row filter {filter_} is not valid.")
        region_length = gr.end - gr.start
        len_ratio_th = self.properties["length_ratio_thresh"]
        df = df[(df["end"] - df["start"]) > region_length * len_ratio_th]
        features = []
        for _, row in df.iterrows():
            gf = GraphicFeature(
                start=row['start'],
                end=row['end'],
                strand=(1 if row['strand'] == '+' else -1),
                label=row['gene_name'],
                color=random.choice(self.colors),
            )
            features.append(gf)
        record = GraphicRecord(
            sequence_length=gr.end - gr.start,
            features=features,
            first_index=gr.start
        )
        record.plot(
            ax=ax,
            with_ruler=False,
            draw_line=False
        )
        self.plot_label()



from .bam import Track
from .bam import BAM
from .gtf import GTF
from .ideogram import Ideogram
from .pseudo import Spacer, HLine, XAxis, ChromName
from .bed import BedBase, BED, TAD  # no all-in class/function
from .hicmat import HicMatBase, Cool, DotHiC, HiCDiff, Selfish, HiCMat
from .hist import HistBase, BedGraph, BigWig, ABCompartment, DiScore, InsuScore, Virtual4C, BAMCov, SNP, Hist
from .arcs import ArcsBase, Pairs, BEDPE, HiCPeaks, Arcs



import logging

from IPython.display import display

from coolbox.utilities import (
    GenomeRange, GenomeLength, BUILT_IN_GENOMES, get_size, fig2bytes
)
from .widgets import SimpleWidgets, FullWidgets

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class Browser(object):
    """
    Browser base class.
    """

    MAX_CACHE_SIZE = 10 ** 7  # max fig cache size.(bytes)

    def __init__(self, frame, reference_genome='hg19',
                 init_range=None, widgets_box='simple',
                 dpi=None, img_format='svg'):
        """
        Parameters
        ----------
        frame : {coolbox.core.frame.Frame, coolbox.core.superframe.base.Superframe}
            Browser's main frame.

        reference_genome : str, optional
            Reference genome,
            built-in references:('hg19', 'hg38', 'mm9', 'mm10')
            if you want use other genome, you can specify the "chromosome length file",
            that is a tab splited file, first column is the chromosomes,
            and second column is the length of correspond chromosome. ['hg19']

        init_range : str, optional
            Initial browser range.

        widgets_box : {'simple', 'full'}, optional
            WidgetsBox sub class, default SimpleWidgets

        dpi : int, optional
            The dpi of frame's image.

        img_format : str, optional
            Frame image format, default svg.
        """

        self.dpi = dpi
        self.img_format = img_format
        self.frame = frame

        if reference_genome in BUILT_IN_GENOMES:
            self.chrom_lengthes = BUILT_IN_GENOMES[reference_genome]
        else:
            self.chrom_lengthes = GenomeLength(reference_genome)
            if len(self.chrom_lengthes) == 0:
                raise IOError("chromosome lengthes file is not include any useful information."
                              "Please check file \"{}\".".format(reference_genome))

        if init_range is not None:
            self.current_range = GenomeRange(init_range)
        else:
            self.current_range = self.get_init_range()

        if widgets_box == 'simple':
            self.widgets = SimpleWidgets(self)
        elif widgets_box == 'full':
            self.widgets = FullWidgets(self)
        else:
            raise NotImplementedError("widgets type {} not support, please use 'simple' or 'full'".format(widgets_box))

        self.goto(self.current_range)
        self.fig = None

        # cache figs in dict, speed up the figure display process.
        #   key: genome range
        #   value: fig image bytes
        self.fig_cache = {}

    def get_init_range(self, chrom=None):
        """
        Generate an initial range within a chromosome.

        Args:
            chrom (str, optional): initial choromosome.

        Return:
            (:obj:`GenomeRange`)
        """

        if chrom is None:
            chrom = list(self.chrom_lengthes.keys())[0]

        default_length = 10 ** 7

        if self.chrom_lengthes[chrom] > default_length:
            range_ = GenomeRange(chrom, 1, default_length)
        else:
            range_ = GenomeRange(chrom, 1, self.chrom_lengthes[chrom])

        return range_

    @property
    def window_size(self):
        size = self.current_range.end - self.current_range.start
        return size

    @property
    def center(self):
        center = (self.current_range.start + self.current_range.end) // 2
        return center

    def goto(self, genome_range, who=None):
        if isinstance(genome_range, str):
            genome_range = GenomeRange(genome_range)
        if not self.chrom_lengthes.check_range(genome_range):
            log.warning("The genome range {} is not valid.".format(genome_range))
            return
        self.current_range = genome_range
        frame_range = GenomeRange(genome_range.chrom,
                                  genome_range.start - 1,  # NOTE: frame's start is zero based
                                  genome_range.end)
        self.frame.goto(frame_range)

        self.widgets.refresh_widgets(who=who)

    def go_left(self, step_ratio=0.5, dry_run=False):
        window_size = self.window_size
        step = int(window_size * step_ratio)
        start = self.current_range.start - step
        end = self.current_range.end - step
        genome_range = GenomeRange(self.current_range.chrom, start, end)
        genome_range = self.chrom_lengthes.bound_range(genome_range)
        if dry_run:
            return genome_range
        else:
            self.goto(genome_range)

    def go_right(self, step_ratio=0.5, dry_run=False):
        window_size = self.window_size
        step = int(window_size * step_ratio)
        start = self.current_range.start + step
        end = self.current_range.end + step
        genome_range = GenomeRange(self.current_range.chrom, start, end)
        genome_range = self.chrom_lengthes.bound_range(genome_range)
        self.goto(genome_range)
        if dry_run:
            return genome_range
        else:
            self.goto(genome_range)

    def zoom_in(self, zoom_ratio=2, dry_run=False):
        window_size = self.window_size
        window_size = window_size // zoom_ratio
        start = self.center - window_size // 2
        end = start + window_size
        genome_range = GenomeRange(self.current_range.chrom, start, end)
        genome_range = self.chrom_lengthes.bound_range(genome_range)
        self.goto(genome_range)
        if dry_run:
            return genome_range
        else:
            self.goto(genome_range)

    def zoom_out(self, zoom_ratio=2, dry_run=False):
        window_size = self.window_size
        window_size = window_size * zoom_ratio
        start = self.center - window_size // 2
        end = start + window_size
        genome_range = GenomeRange(self.current_range.chrom, start, end)
        genome_range = self.chrom_lengthes.bound_range(genome_range)
        self.goto(genome_range)
        if dry_run:
            return genome_range
        else:
            self.goto(genome_range)

    def show(self, *args, **kwargs):
        """
        Show widgets and frame.
        """
        display(self.widgets.panel)
        self.refresh()

    def refresh(self, hard=False):
        """
        Refresh the image display.
        """
        if (self.current_range in self.fig_cache) and (not hard):
            fig_bytes = self.fig_cache[self.current_range]
        else:
            fig_current = self.frame.show()
            fig_bytes = fig2bytes(fig_current, encode=self.img_format, dpi=self.dpi)
            if self.img_format == 'svg':
                fig_bytes = fig_bytes.decode("utf-8")
            self.fig_cache[self.current_range] = fig_bytes

        # auto clear fig cache for prevent memory leak
        if len(self.fig_cache) > 20 and \
                get_size(self.fig_cache) >= Browser.MAX_CACHE_SIZE:
            self.clear_fig_cache()

        self.widgets.frame_widget.value = fig_bytes

    def preload_imgs(self, directions):
        """
        Preloading images to self.fig_cache.

        Can load image in one of 4 directions:
            left, right, zoom-in, zoom-out
        or load all directions.
        """

        def preload(g_range):
            if g_range not in self.fig_cache:
                fig = self.frame.plot(g_range.chrom, g_range.start, g_range.end)
                self.fig_cache[g_range] = fig2bytes(fig, encode=self.img_format, dpi=self.dpi)

        possible_dires = ['left', 'right', 'zoom-in', 'zoom-out']

        if isinstance(directions, str):
            if directions == 'all':
                all_directions = possible_dires
            else:
                all_directions = [directions]
        else:
            assert isinstance(directions, list) or isinstance(directions, tuple), \
                "direction must str or list or tuple object."
            all_directions = directions

        for dire in all_directions:
            assert directions in possible_dires, \
                "direction must one of {}.".format(possible_dires)
            if dire == 'left':
                left_region = self.go_left(dry_run=True)
                preload(left_region)
            elif dire == 'right':
                right_region = self.go_right(dry_run=True)
                preload(right_region)
            elif dire == 'zoom-in':
                zoom_in_region = self.zoom_in(dry_run=True)
                preload(zoom_in_region)
            elif dire == 'zoom-out':
                zoom_out_region = self.zoom_out(dry_run=True)
                preload(zoom_out_region)

    def clear_fig_cache(self):
        """
        Clear the fig cache.
        """
        self.fig_cache = {}

    def save(self, path, dpi=None):
        """
        Save current frame's image to file.
        """
        c_fig = self.frame.show()
        dpi = dpi or self.dpi
        c_fig.savefig(path, dpi=dpi)

    @property
    def tracks(self):
        return self.frame.tracks

    def fetch_data(self, genome_range=None):
        return self.frame.fetch_data(genome_range)



from typing import Union
import matplotlib
import matplotlib.colors as colors
import numpy as np
import pandas as pd
from matplotlib.patches import Rectangle

from coolbox.utilities import (
    get_logger
)
from coolbox.utilities.genome import GenomeRange

log = get_logger(__name__)


class PlotBed(object):
    def init_for_plot(self):
        from matplotlib import font_manager

        properties = self.properties
        self.len_w = None  # this is the length of the letter 'w' given the font size
        self.counter = None
        self.small_relative = None
        self.is_draw_labels = properties['labels'] == 'on'
        self.fp = font_manager.FontProperties(size=properties['fontsize'])
        # set the distance between rows
        self.row_scale = properties['interval_height'] * 2.3
        # set colormap if possible, else None, wait for setting when plot
        self.colormap = None
        if not matplotlib.colors.is_color_like(self.properties['color']) and self.properties['color'] != 'bed_rgb':
            # check if the color is a valid colormap name
            if self.properties['color'] not in matplotlib.cm.datad:
                log.warning("*WARNING* color: '{}' for Track {} is not valid. Color has "
                            "been set to {}".format(self.properties['color'], self.properties['name'],
                                                    self.COLOR))
                self.properties['color'] = self.COLOR
            else:
                self.colormap = self.properties['color']

    def plot_genes(self, ax, gr: GenomeRange, ov_genes: pd.DataFrame):
        properties = self.properties
        # bed_type
        self.properties['bed_type'] = properties['bed_type'] or self.infer_bed_type(ov_genes)
        # as min_score and max_score change every plot, we compute them for every plot
        min_score, max_score = properties['min_score'], properties['max_score']
        has_score_col = properties['bed_type'] in ('bed6', 'bed9', 'bed12')
        if has_score_col and len(ov_genes):
            min_score = (min_score != 'inf') or ov_genes['score'].min()
            max_score = (max_score != '-inf') or ov_genes['score'].max()
        min_score, max_score = float(min_score), float(max_score)

        # set colormap
        if self.colormap is not None:
            norm = matplotlib.colors.Normalize(vmin=min_score, vmax=max_score)
            cmap = matplotlib.cm.get_cmap(properties['color'])
            self.colormap = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap)
        if properties['color'] == 'bed_rgb' and properties['bed_type'] not in ['bed12', 'bed9']:
            log.warning("*WARNING* Color set to 'bed_rgb', but bed file does not have the rgb field. The color has "
                        "been set to {}".format(self.COLOR))
            self.properties['color'] = self.COLOR
            self.colormap = None


        self.counter = 0
        self.small_relative = 0.004 * (gr.end - gr.start)
        self.get_length_w(ax.get_figure().get_figwidth(), gr.start, gr.end)
        # turn labels off when too many intervals are visible.
        if properties['labels'] == 'on' and len(ov_genes) > 60:
            self.is_draw_labels = False

        num_rows = properties['num_rows']
        max_num_row_local = 1
        max_ypos = 0
        # check for the number of other intervals that overlap
        #    with the given interval
        #            1         2
        #  012345678901234567890123456
        #  1=========       4=========
        #       2=========
        #         3============
        #
        # for 1 row_last_position = [9]
        # for 2 row_last_position = [9, 14]
        # for 3 row_last_position = [9, 14, 19]
        # for 4 row_last_position = [26, 14, 19]

        row_last_position = []  # each entry in this list contains the end position
        # of genomic interval. The list index is the row
        # in which the genomic interval was plotted.
        # Any new genomic interval that wants to be plotted,
        # knows the row to use by finding the list index that
        # is larger than its start

        # check for overlapping genes including
        # label size (if plotted)

        for bed in ov_genes.itertuples():
            """
            BED12 gene format with exon locations at the end
            chrX    20850   23076   CG17636-RA      0       -       20850   23017   0       3       946,765,64,     0,1031,2162,

            BED9
            bed with rgb at end
            chr2L   0       70000   ID_5    0.26864549832   .       0       70000   51,160,44

            BED6
            bed without rgb
            chr2L   0       70000   ID_5    0.26864549832   .

            BED3
            bed with only intervals
            chr2L  0        70000
            """
            self.counter += 1

            if self.is_draw_labels:
                num_name_characters = len(bed.name) + 2  # +2 to account for an space before and after the name
                bed_extended_end = int(bed.end + (num_name_characters * self.len_w))
            else:
                bed_extended_end = (bed.end + 2 * self.small_relative)

            # get smallest free row
            if len(row_last_position) == 0:
                free_row = 0
                row_last_position.append(bed_extended_end)
            else:
                # get list of rows that are less than bed.start, then take the min
                idx_list = [idx for idx, value in enumerate(row_last_position) if value < bed.start]
                if len(idx_list):
                    free_row = min(idx_list)
                    row_last_position[free_row] = bed_extended_end
                else:
                    free_row = len(row_last_position)
                    row_last_position.append(bed_extended_end)

            rgb, edgecolor = self.get_rgb_and_edge_color(bed)

            ypos = self.get_y_pos(free_row)

            # do not plot if the maximum interval rows to plot is reached
            if num_rows and free_row >= float(num_rows):
                continue

            if free_row > max_num_row_local:
                max_num_row_local = free_row
            if ypos > max_ypos:
                max_ypos = ypos

            if properties['bed_type'] == 'bed12':
                if properties['gene_style'] == 'flybase':
                    self.draw_gene_with_introns_flybase_style(ax, bed, ypos, rgb, edgecolor)
                else:
                    self.draw_gene_with_introns(ax, bed, ypos, rgb, edgecolor)
            else:
                self.draw_gene_simple(ax, bed, ypos, rgb, edgecolor)

            if self.is_draw_labels and bed.start > gr.start and bed.end < gr.end:
                ax.text(bed.end + self.small_relative,
                        ypos + (float(properties['interval_height']) / 2),
                        bed.name,
                        horizontalalignment='left',
                        verticalalignment='center',
                        fontproperties=self.fp)

        if self.counter == 0:
            log.warning(f"*Warning* No intervals were found for file {properties['file']} "
                        f"in Track \'{properties['name']}\' for the interval plotted ({gr}).\n")

        ymax = 0
        if num_rows:
            ymin = float(num_rows) * self.row_scale
        else:
            ymin = max_ypos + properties['interval_height']

        log.debug("ylim {},{}".format(ymin, ymax))
        # the axis is inverted (thus, ymax < ymin)
        ax.set_ylim(ymin, ymax)

        if properties['display'] == 'domain':
            ax.set_ylim(-5, 205)
        elif properties['display'] == 'collapsed':
            ax.set_ylim(-5, 105)

        ax.set_xlim(gr.start, gr.end)

    def draw_gene_with_introns(self, ax, bed, ypos, rgb, edgecolor):
        """
        draws a gene like in flybase gbrowse.
        """
        from matplotlib.patches import Polygon
        properties = self.properties
        height = float(properties['interval_height'])

        if bed.block_count == 0 and bed.thick_start == bed.start and bed.thick_end == bed.end:
            self.draw_gene_simple(ax, bed, ypos, rgb, edgecolor)
            return
        half_height = height / 2
        quarter_height = height / 4
        three_quarter_height = quarter_height * 3

        # draw 'backbone', a line from the start until the end of the gene
        ax.plot([bed.start, bed.end], [ypos + half_height, ypos + half_height], 'black', linewidth=0.5, zorder=-1)

        for idx in range(0, bed.block_count):
            x0 = bed.start + bed.block_starts[idx]
            x1 = x0 + bed.block_sizes[idx]
            if x1 < bed.thick_start or x0 > bed.thick_end:
                y0 = ypos + quarter_height
                y1 = ypos + three_quarter_height
            else:
                y0 = ypos
                y1 = ypos + height

            if x0 < bed.thick_start < x1:
                vertices = ([(x0, ypos + quarter_height), (x0, ypos + three_quarter_height),
                             (bed.thick_start, ypos + three_quarter_height),
                             (bed.thick_start, ypos + height),
                             (bed.thick_start, ypos + height),
                             (x1, ypos + height), (x1, ypos),
                             (bed.thick_start, ypos), (bed.thick_start, ypos + quarter_height)])

            elif x0 < bed.thick_end < x1:
                vertices = ([(x0, ypos),
                             (x0, ypos + height),
                             (bed.thick_end, ypos + height),
                             (bed.thick_end, ypos + three_quarter_height),
                             (x1, ypos + three_quarter_height),
                             (x1, ypos + quarter_height),
                             (bed.thick_end, ypos + quarter_height),
                             (bed.thick_end, ypos)])
            else:
                vertices = ([(x0, y0), (x0, y1), (x1, y1), (x1, y0)])

            ax.add_patch(Polygon(vertices, closed=True, fill=True,
                                 linewidth=0.1,
                                 edgecolor='none',
                                 facecolor=rgb))

            if idx < bed.block_count - 1:
                # plot small arrows using the character '<' or '>' over the back bone
                intron_length = bed.block_starts[idx + 1] - (bed.block_starts[idx] + bed.block_sizes[idx])
                marker = 5 if bed.strand == '+' else 4
                if intron_length > 3 * self.small_relative:
                    pos = np.arange(x1 + 1 * self.small_relative,
                                    x1 + intron_length + self.small_relative, int(2 * self.small_relative))
                    ax.plot(pos, np.zeros(len(pos)) + ypos + half_height, '.', marker=marker,
                            fillstyle='none', color='blue', markersize=3)

                elif intron_length > self.small_relative:
                    intron_center = x1 + int(intron_length) / 2
                    ax.plot([intron_center], [ypos + half_height], '.', marker=5,
                            fillstyle='none', color='blue', markersize=3)

    def draw_gene_with_introns_flybase_style(self, ax, bed, ypos, rgb, edgecolor):
        """
        draws a gene using different styles
        """
        from matplotlib.patches import Polygon
        properties = self.properties
        height = float(properties['interval_height'])

        if bed.block_count == 0 and bed.thick_start == bed.start and bed.thick_end == bed.end:
            self.draw_gene_simple(ax, bed, ypos, rgb, edgecolor)
            return
        half_height = height / 2
        # draw 'backbone', a line from the start until the end of the gene
        ax.plot([bed.start, bed.end], [ypos + half_height, ypos + half_height], 'black', linewidth=0.5, zorder=-1)

        # get start, end of all the blocks
        positions = []
        for idx in range(0, bed.block_count):
            x0 = bed.start + bed.block_starts[idx]
            x1 = x0 + bed.block_sizes[idx]
            if x0 < bed.thick_start < x1:
                positions.append((x0, bed.thick_start, 'UTR'))
                positions.append((bed.thick_start, x1, 'coding'))

            elif x0 < bed.thick_end < x1:
                positions.append((x0, bed.thick_end, 'coding'))
                positions.append((bed.thick_end, x1, 'UTR'))

            else:
                if x1 < bed.thick_start or x0 > bed.thick_end:
                    seg_type = 'UTR'
                else:
                    seg_type = 'coding'

                positions.append((x0, x1, seg_type))

        # plot all blocks as rectangles except the last if the strand is + or
        # the first is the strand is -, which are drawn as arrows.
        if bed.strand == '-':
            positions = positions[::-1]

        first_pos = positions.pop()
        if first_pos[2] == 'UTR':
            _rgb = 'grey'
        else:
            _rgb = rgb

        vertices = self.draw_arrow(ax, first_pos[0], first_pos[1], bed.strand, ypos)

        ax.add_patch(Polygon(vertices, closed=True, fill=True,
                             edgecolor=edgecolor,
                             facecolor=_rgb,
                             linewidth=0.5))

        for start_pos, end_pos, _type in positions:
            if _type == 'UTR':
                _rgb = 'grey'
            else:
                _rgb = rgb
            vertices = [(start_pos, ypos), (start_pos, ypos + height),
                        (end_pos, ypos + height), (end_pos, ypos)]

            ax.add_patch(Polygon(vertices, closed=True, fill=True,
                                 edgecolor=edgecolor,
                                 facecolor=_rgb,
                                 linewidth=0.5))

    def draw_gene_simple(self, ax, bed, ypos, rgb, edgecolor):
        """
        draws an interval with direction (if given)
        """
        from matplotlib.patches import Polygon
        properties = self.properties
        height = float(properties['interval_height'])

        if bed.strand not in ['+', '-']:
            ax.add_patch(Rectangle((bed.start, ypos), bed.end - bed.start, height,
                                   edgecolor=edgecolor, facecolor=rgb, linewidth=0.5))
        else:
            vertices = self.draw_arrow(ax, bed.start, bed.end, bed.strand, ypos)
            ax.add_patch(Polygon(vertices, closed=True, fill=True,
                                 edgecolor=edgecolor,
                                 facecolor=rgb,
                                 linewidth=0.5))

    def draw_arrow(self, ax, start, end, strand, ypos):
        properties = self.properties
        height = float(properties['interval_height'])

        half_height = height / 2
        if strand == '+':
            x0 = start
            x1 = end  # - self.small_relative
            y0 = ypos
            y1 = ypos + height

            """
            The vertices correspond to 5 points along the path of a form like the following,
            starting in the lower left corner and progressing in a clock wise manner.

            -----------------\
            ---------------- /

            """

            vertices = [(x0, y0), (x0, y1), (x1, y1), (x1 + self.small_relative, y0 + half_height), (x1, y0)]

        else:
            x0 = start  # + self.small_relative
            x1 = end
            y0 = ypos
            y1 = ypos + height

            """
            The vertices correspond to 5 points along the path of a form like the following,
            starting in the lower left corner and progressing in a clock wise manner.

            /-----------------
            \\-----------------
            """

            vertices = [(x0, y0), (x0 - self.small_relative, y0 + half_height), (x0, y1), (x1, y1), (x1, y0)]

        return vertices

    def get_length_w(self, fig_width, region_start, region_end):
        """
        to improve the visualization of the genes it is good to have an estimation of the label
        length. In the following code I try to get the length of a 'W' in base pairs.
        """
        if self.is_draw_labels:
            # from http://scipy-cookbook.readthedocs.org/items/Matplotlib_LaTeX_Examples.html
            inches_per_pt = 1.0 / 72.27
            font_in_inches = self.properties['fontsize'] * inches_per_pt
            region_len = region_end - region_start
            bp_per_inch = region_len / fig_width
            font_in_bp = font_in_inches * bp_per_inch
            self.len_w = font_in_bp
            log.debug("len of w set to: {} bp".format(self.len_w))
        else:
            self.len_w = 1

        return self.len_w

    def get_y_pos(self, free_row):
        """
        The y_pos is set such that regions to be plotted do not overlap (stacked). To override this
        the properties['collapsed'] needs to be set.

        The algorithm uses a interval tree (self.region_interval) to check the overlaps
        and a sort of coverage vector 'rows used' to identify the row in which to plot

        Return
        ------
        ypos : int
            y position.
        """

        # if the domain directive is given, ypos simply oscilates between 0 and 100
        if self.properties['display'] == 'interlaced':
            ypos = self.properties['interval_height'] if self.counter % 2 == 0 else 1
        elif self.properties['display'] == 'collapsed':
            ypos = 0
        else:
            ypos = free_row * self.row_scale
        return ypos

    def get_rgb_and_edge_color(self, bed):
        # TODO need simplification
        rgb = self.properties['color']
        edgecolor = self.properties['border_color']

        if self.colormap:
            # translate value field (in the example above is 0 or 0.2686...) into a color
            rgb = self.colormap.to_rgba(bed.score)

        # for tad coverage
        if self.properties['style'] == 'tad' and self.properties['border_only'] == 'yes':
            rgb = 'none'
        elif self.properties['color'] == 'bed_rgb':
            # if rgb is set in the bed line, this overrides the previously
            # defined colormap
            if self.properties['bed_type'] in ['bed9', 'bed12'] and len(bed.rgb) == 3:
                try:
                    rgb = [float(x) / 255 for x in bed.rgb]
                    if 'border_color' in self.properties:
                        edgecolor = self.properties['border_color']
                    else:
                        edgecolor = self.properties['color']
                except IndexError:
                    rgb = self.COLOR
            else:
                rgb = self.COLOR
        return rgb, edgecolor

    @staticmethod
    def infer_bed_type(df: pd.DataFrame) -> Union[str, None]:
        #  bed_type of dataframe are store in dataframe's __dict__ in FetchBed.fetch_intervals
        if 'bed_type' in df.__dict__:
            bed_type = df.bed_type
        else:
            bed_types = {
                12: 'bed12',
                9: 'bed9',
                6: 'bed6',
                3: 'bed3'
            }
            num_col = len(df.columns)
            bed_type = bed_types[num_col] if num_col in bed_types else 'bed3'
            if bed_type == 'bed3' and num_col < 3:
                raise ValueError(f"Invalid dataframe for bed3 with columns: {df.columns}")
        return bed_type

    def plot_tads(self, ax, gr: GenomeRange, tads: pd.DataFrame):
        """
        Plots the boundaries as triangles in the given ax.
        """
        from coolbox.core.track.hicmat import HicMatBase
        # coverage only
        assert 'track' in self.__dict__ and isinstance(self.track, HicMatBase), \
            f"The parent track should be instance of {HicMatBase}"

        hicmat_tri_style = (HicMatBase.STYLE_WINDOW, HicMatBase.STYLE_TRIANGULAR)
        hicmat_ma_style = (HicMatBase.STYLE_MATRIX,)

        hictrack = self.track
        hicmat_style = hictrack.properties['style']

        # TODO Should we add plotting in BigWig, BedGraph, ABCCompartment, Arcs support?(The original codes supports)
        for region in tads.itertuples():
            if hicmat_style in hicmat_tri_style:
                depth = (gr.end - gr.start) / 2
                ymax = (gr.end - gr.start)
                self.plot_triangular(ax, gr, region, ymax, depth)
            elif hicmat_style in hicmat_ma_style:
                self.plot_box(ax, gr, region)
            else:
                raise ValueError(f"unsupported hicmat style {hicmat_style}")

        if len(tads) == 0:
            log.warning("No regions found for Coverage {}.".format(self.properties['name']))

    def plot_triangular(self, ax, gr, region, ymax, depth):
        """
              /\
             /  \
            /    \
        _____________________
           x1 x2 x3
        """

        from matplotlib.patches import Polygon
        x1 = region.start
        x2 = x1 + float(region.end - region.start) / 2
        x3 = region.end
        y1 = 0
        y2 = (region.end - region.start)

        y = (y2 / ymax) * depth

        rgb, edgecolor = self.get_rgb_and_edge_color(region)

        triangle = Polygon(np.array([[x1, y1], [x2, y], [x3, y1]]), closed=True,
                           facecolor=rgb, edgecolor=edgecolor,
                           alpha=self.properties['alpha'],
                           linestyle=self.properties['border_style'],
                           linewidth=self.properties['border_width'])
        ax.add_artist(triangle)
        self.plot_score(ax, gr, region, 'triangular', ymax, depth)

    def plot_box(self, ax, gr, region):
        from matplotlib.patches import Rectangle

        x1 = region.start
        x2 = region.end
        x = y = x1
        w = h = (x2 - x1)

        rgb, edgecolor = self.get_rgb_and_edge_color(region)

        fill = True if self.properties['border_only'] == 'no' else False

        rec = Rectangle((x, y), w, h,
                        fill=fill,
                        facecolor=rgb,
                        edgecolor=edgecolor,
                        alpha=self.properties['alpha'],
                        linestyle=self.properties['border_style'],
                        linewidth=self.properties['border_width'])
        ax.add_patch(rec)
        self.plot_score(ax, gr, region, 'box')

    def plot_score(self, ax, gr, region, style, ymax=None, depth=None):
        properties = self.properties

        if properties['show_score'] != 'yes':
            return
        bed = region
        score = bed.score
        if not (isinstance(score, float) or isinstance(score, int)):
            # score is not number not plot
            return
        region_length = region.end - region.start
        if region_length / gr.length < 0.05:
            # region too small not plot score
            return
        font_size = properties['score_font_size']
        if font_size == 'auto':
            # inference the font size
            from math import log2
            base_size = 18
            s_ = (region_length / gr.length) * 10
            s_ = int(log2(s_))
            font_size = base_size + s_
        ratio = properties['score_height_ratio']
        color = properties['score_font_color']
        if style == 'box':
            x1 = region.start
            x2 = region.end
            w = x2 - x1
            x = x2 - w * ratio
            y = x1 + w * ratio
        else:  # triangular
            x = region.begin + region_length * 0.4
            y = (region_length / ymax) * depth * ratio
        ax.text(x, y, "{0:.3f}".format(score), fontsize=font_size, color=color)



import numpy as np
import pandas as pd
from dna_features_viewer import GraphicFeature, GraphicRecord

from coolbox.utilities import (
    get_logger, GenomeRange, split_genome_range
)
from coolbox.utilities.bam import process_bam, query_bam
from .base import Track

log = get_logger(__name__)


class BAM(Track):
    """
    BAM alignment track for plotting reads.

    Parameters
    ----------
    file : str
        Path to .bam .sam file.

    length_ratio_thresh : float
        Length ratio threshold of show alignments. (Default 0.01)

    """

    DEFAULT_PROPERTIES = {
        "height": 3,
        "length_ratio_thresh": 0.05,
        "color": "#6688ff"
    }

    def __init__(self, file, **kwargs):
        properties = BAM.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs,
        })
        properties.update(kwargs)
        super().__init__(properties)
        self.indexed_bam = process_bam(file)

    def fetch_data(self, gr: GenomeRange, **kwargs) -> pd.DataFrame:
        """

        Returns
        -------
        intervals : pandas.core.frame.DataFrame
            Sam interval table.
            The DataFrame table should has columns like:

            columns = ["qname", "flag", "rname", "pos", "mapq", "cigar",
                      "rnext", "pnext", "tlen", "seq", "qual", "options"]
        """
        return self.fetch_intervals(gr)

    def plot(self, ax, gr: GenomeRange, **kwargs):
        self.plot_align(ax, gr)

    def plot_align(self, ax, gr: GenomeRange):
        assert isinstance(gr, GenomeRange), "The input gr should be type GenomeRange"
        df = self.fetch_plot_data(gr)
        df_ = df[np.bitwise_and(df['flag'], 0b100) == 0]
        len_thresh = self.properties["length_ratio_thresh"]
        df_ = df_[df_['seq'].str.len() > (gr.length * len_thresh)]
        if df_.shape[0] <= 0:
            return
        rev_flag = np.bitwise_and(df['flag'], 0b10000) != 0
        features = []
        for idx, row in df_.iterrows():
            start = row['pos'] - gr.start
            end = row['pos'] + len(row['seq']) - gr.start
            strand = -1 if rev_flag.iloc[idx] else 1
            gf = GraphicFeature(
                start=start,
                end=end,
                strand=strand,
                color=self.properties['color'],
            )
            features.append(gf)
        record = GraphicRecord(sequence_length=gr.length, features=features)
        record.plot(
            ax=ax,
            with_ruler=False,
            draw_line=False
        )

    def fetch_intervals(self, genome_range: GenomeRange):
        chrom, start, end = split_genome_range(genome_range)
        rows = []
        for row_items in query_bam(self.indexed_bam, chrom, start, end, split=True):
            rows.append(row_items)
        # https://samtools.github.io/hts-specs/SAMv1.pdf
        fields = ["qname", "flag", "rname", "pos", "mapq", "cigar",
                  "rnext", "pnext", "tlen", "seq", "qual", "options"]
        df = pd.DataFrame(rows, columns=fields)
        if df.shape[0] > 0:
            df['flag'] = df['flag'].astype(int)
            df['pos'] = df['pos'].astype(int)
            df['mapq'] = df['mapq'].astype(int)
        return df



import logging

from ipywidgets import HTML

from .base import WidgetsBox
from .navigation import NavigationBar

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class SimpleWidgets(WidgetsBox):
    """
    Simplest widgets panel design.

    -----

    """

    def __init__(self, browser, *args, **kwargs):
        frame_widget = HTML()
        self.navigation_bar = NavigationBar(browser, frame_widget)
        super().__init__(browser, frame_widget, *args, **kwargs)

    def get_widgets_dict(self):
        widgets = self.navigation_bar.widgets
        return widgets

    def refresh_widgets(self, who=None):
        self.navigation_bar.refresh_widgets(self.browser, who)

    def compose_panel(self, widgets_dict):
        panel = self.navigation_bar.panel
        return panel

    def register_events_handler(self):
        self.navigation_bar.register_events_handler(self.browser)



from typing import Tuple, Sequence, Union
import pandas as pd

from coolbox.utilities import (
    get_logger
)
from coolbox.utilities.bed import (
    ReadBed, query_bed
)

from coolbox.utilities.genome import GenomeRange

log = get_logger(__name__)


class FetchBed(object):
    def fetch_intervals(self, bgz_file, gr: GenomeRange) -> pd.DataFrame:
        """
        Fetch intervals within input chromosome range.
        """
        intervals, bed_type = self.load_range(bgz_file, gr)
        if len(intervals) == 0:
            gr.change_chrom_names()
            intervals, bed_type = self.load_range(bgz_file, gr)
            if len(intervals) == 0:
                log.debug(f"No valid intervals were found in file {bgz_file} within range {gr}")
        intval_table = self.intervals2dataframe(intervals, bed_type)
        # attach bed_type info onto the dataframe
        intval_table.bed_type = bed_type
        return intval_table

    @staticmethod
    def load_range(bgz_file, gr: GenomeRange) -> Tuple[Sequence, Union[str, None]]:
        intervals = []
        try:
            bed_iterator = ReadBed(query_bed(bgz_file, gr.chrom, gr.start, gr.end))
        except StopIteration:
            log.info(f"No records in the range {str(gr)}")
            return [], None

        for bed in bed_iterator:
            intervals.append(bed)

        return intervals, bed_iterator.file_type

    @staticmethod
    def intervals2dataframe(intervals, bed_type: Union[str, None]) -> pd.DataFrame:
        """
        Convert intervals list to pandas.DataFrame
        """
        bed_fields = ['chromosome', 'start', 'end',
                      'name', 'score', 'strand',
                      'thick_start', 'thick_end',
                      'rgb', 'block_count',
                      'block_sizes', 'block_starts']

        if bed_type == 'bed12':
            fields = bed_fields
        elif bed_type == 'bed9':
            fields = bed_fields[:9]
        else:
            fields = bed_fields[:6]

        df = pd.DataFrame(intervals, columns=fields)

        return df



from copy import copy

from coolbox.utilities import op_err_msg, get_feature_stack

__all__ = [
    "Feature", "Color", "ColorMap",
    "TrackHeight", "Inverted", "Title",
    "MaxValue", "MinValue", "HistStyle", "ShowDataRange",
    "ShowColorBar", "DepthRatio", "CoolStyle",
    "FrameFeature", "FrameTitle"
]


class Feature(object):
    """
    Feature base class.
    """

    def __init__(self, **kwargs):
        self.properties = kwargs

    def __add__(self, other):
        from .track.base import Track
        from .frame.base import FrameBase
        from .coverage.base import Coverage
        from .coverage.base import CoverageStack

        if isinstance(other, Track):
            result = copy(other)
            result.properties.update(self.properties)
            return result
        elif isinstance(other, FrameBase):
            result = copy(other)
            if len(result.tracks) != 0:
                first = list(result.tracks.values())[0]
                first.properties.update(self.properties)
            return result
        elif isinstance(other, Coverage):
            result = copy(other)
            result.properties.update(self.properties)
            return result
        elif isinstance(other, CoverageStack):
            result = copy(other)
            if len(result.coverages) != 0:
                first = result.coverages[0]
                first.properties.update(self.properties)
            return result
        else:
            raise TypeError(op_err_msg(self, other))

    def __mul__(self, other):
        from .frame.base import FrameBase

        if isinstance(other, FrameBase):
            result = copy(other)
            result.add_feature_to_tracks(self)
            return result
        else:
            raise TypeError(op_err_msg(self, other, op='*'))

    def __enter__(self):
        stack = get_feature_stack()
        stack.append(self)
        return self

    def __exit__(self, type, value, traceback):
        stack = get_feature_stack()
        stack.pop()


class Color(Feature):
    """
    Track color.
    """

    def __init__(self, value):
        super().__init__(color=value)


class ColorMap(Feature):
    """
    Track color map.
    """

    def __init__(self, value):
        # TODO which one?
        super().__init__(color=value, cmap=value)


class TrackHeight(Feature):
    """
    Track height.
    """

    def __init__(self, value):
        super().__init__(height=value)


class Inverted(Feature):
    """
    Invert the orientation of track.
    """

    def __init__(self):
        super().__init__(orientation='inverted')


class Title(Feature):
    """
    Track title.
    """

    def __init__(self, value):
        super().__init__(title=value)


class MaxValue(Feature):
    """
    Max value of track.
    """

    def __init__(self, value):
        super().__init__(max_value=value)


class MinValue(Feature):
    """
    Min value of track.
    """

    def __init__(self, value):
        super().__init__(min_value=value)


class HistStyle(Feature):
    """
    Style of BigWig or BedGraph.
    """

    def __init__(self, style='fill', fmt="-", size=10, line_width=2.0):
        from coolbox.core.track import HistBase
        assert style in HistBase.STYLES, f"The style should be one of {HistBase.STYLES}"
        super().__init__(style=style, fmt=fmt, size=size, line_width=line_width)


class ShowDataRange(Feature):
    """
    Show data range or not.
    """

    def __init__(self, data_range_style="y-axis"):
        assert data_range_style in ("no", False, "text", "y-axis"), "Should be one of ['no', 'text', 'y-axis']"
        super().__init__(data_range_style=data_range_style)


class ShowColorBar(Feature):
    """
    Show color bar or not.
    """

    def __init__(self, show=True):
        super().__init__(color_bar='yes' if bool(show) else 'no')


class DepthRatio(Feature):
    """
    Control Cool track's depth ratio.
    """

    def __init__(self, depth_ratio):
        super().__init__(depth_ratio=depth_ratio)


class CoolStyle(Feature):
    """
    Control Cool track's style.
    """

    def __init__(self, style='triangular'):
        super().__init__(style=style)


class FrameFeature(Feature):
    """
    FrameFeature base class.
    """

    def __add__(self, other):
        from .frame.base import FrameBase

        if isinstance(other, FrameBase):
            properties = self.properties.copy()
            properties.update(other.properties)
            result = copy(other)
            result.properties = properties
            return result
        else:
            raise TypeError(op_err_msg(self, other))


class FrameTitle(FrameFeature):
    """
    Frame title.
    """

    def __init__(self, value):
        super().__init__(title=value)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from .base import Browser
from coolbox.utilities import op_err_msg

__all__ = ["WidgetsPanel", "Browser"]


class WidgetsPanel(object):
    """
    Widgets container.

    Parameters
    ----------
    type : {'simple', 'full'}, optional
        Widgets type 'simple' or 'full',
        default 'simple'

    reference_genome : str, optional
        Reference genome,
        built-in references:('hg19', 'hg38', 'mm9', 'mm10')
        if you want use other genome, you can specify the "chromosome length file",
        that is a tab splited file, first column is the chromosomes,
        and second column is the length of correspond chromosome.
        default 'hg19'

    """

    def __init__(self, type="simple", reference_genome="hg19"):
        self.ref = reference_genome
        self.type = type

    def __add__(self, other):
        """

        >>>

        """
        from ..frame.base import FrameBase
        if isinstance(other, FrameBase):
            return Browser(other, reference_genome=self.ref, widgets_box=self.type)
        else:
            raise TypeError(op_err_msg(self, other, op="+"))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import subprocess as subp
import os.path as osp

import numpy as np


def is_bam_sorted(bam_path):
    p = subp.Popen(['samtools', 'view', '-H', bam_path], stdout=subp.PIPE)
    for line in p.stdout:
        line = line.decode("utf-8")
        if "SO:unsorted" in line:
            return False
    return True


def process_bam(bam_path):
    if bam_path.endswith(".bam"):
        bai_path = bam_path + '.bai'
        if osp.exists(bai_path):
            return bam_path
        if not is_bam_sorted(bam_path):
            sorted_bam_path = bam_path[:-4] + '.sorted.bam'
            subp.check_call(['samtools', 'sort', bam_path, '-o', sorted_bam_path])
        else:
            sorted_bam_path = bam_path
        subp.check_call(['samtools', 'index', sorted_bam_path])
    elif bam_path.endswith(".sam"):
        sorted_bam_path = bam_path[:-4] + '.sorted.bam'
        subp.check_call(['samtools', 'sort', bam_path, '-o', sorted_bam_path])
        subp.check_call(['samtools', 'index', sorted_bam_path])
    else:
        raise IOError("BAM input file should be in .bam or .sam format")
    return sorted_bam_path


def query_bam(filename, chrom, start, end, split=True):
    """Call tabix and generate an array of strings for each line it returns."""
    query = '{}:{}-{}'.format(chrom, start, end)
    p = subp.Popen(['samtools', 'view', filename, query], stdout=subp.PIPE)
    for line in p.stdout:
        line = line.decode('utf-8')
        if not split:
            yield line
        else:
            items = line.strip().split('\t')
            items_ = items[:11] + ["\t".join(items[12:])]
            yield items_


def coverage_by_samtools(bam_path, region, bins):
    cmd = ["samtools", "coverage", bam_path, "-r", region, "-w", str(bins)]
    p = subp.Popen(cmd, stdout=subp.PIPE)
    lines = []
    for line in p.stdout:
        line = line.decode('utf-8')
        lines.append(line)
    covs = parse_samtools_cov(lines)
    return np.array(covs)


def parse_samtools_cov(lines):
    covs = {}
    for line in lines[1:-1]:
        left, mid, _ = line.split("│")
        percent = float(left.strip("> %"))
        for i, c in enumerate(mid):
            covs.setdefault(i, 0)
            if c != ' ' and covs[i] == 0:
                covs[i] = percent
    covs = [covs[i] for i in sorted(covs.keys())]
    return covs



from collections import OrderedDict

from ipywidgets import (
    HBox, VBox, Dropdown, Button, Label,
    Checkbox, FloatText,
    Text, IntRangeSlider, Layout, HTML
)

from coolbox.utilities import (
    GenomeRange
)

ALL_BW_MARK = "ALL(BW/BG)"


class NavigationBar(object):
    """
    Layout:

    -----------------------------------------------------------------------------------------
    left_button | right_button | zoom_out_button | zoom_in_button | range_textbox | go_button
    -----------------------------------------------------------------------------------------
    range_min_label | range_slider | range_max_label
    --------------------------------------------------------------------
    auto_check_box | track_min_val_float_text | track_max_val_float_text
    --------------------------------------------------------------------
    frame
    """

    def __init__(self, browser, frame=None):
        chromsomes = list(browser.chrom_lengthes.keys())
        self.widgets = self.__get_widgets(chromsomes, browser, frame)
        self.panel = self.__compose_panel(self.widgets)
        self.selected_tracks = self.__get_tracks_name(browser)

    def __get_tracks_name(self, browser):
        return [t.name for t in browser.tracks.values() if 'min_value' in t.properties]

    def refresh_widgets(self, browser, who=None):
        range_ = browser.current_range
        max_ = browser.chrom_lengthes[range_.chrom]
        # update range_textbox
        if who != "go_button":
            self.widgets['range_textbox'].value = str(range_)
        # update range_slider
        if who != "range_slider":
            slider = self.widgets['range_slider']
            slider.min = 1
            slider.max = max_
            slider.value = (range_.start, range_.end)
        # update chromosome list
        if who != "chromosomes_list":
            self.widgets['chromosomes_list'].value = range_.chrom
        # update range max min label
        self.widgets['range_min_label'].value = str(1)
        self.widgets['range_max_label'].value = str(max_)

    def register_events_handler(self, browser):
        """
        Parameters
        ----------
        browser : `BrowserBase`
            Browser object.
        """

        # chromosomes_list value change
        def chrom_dropdown_val_change(change):
            new_chrom = change['new']
            current_range = browser.current_range
            # only change chromosome
            range_ = GenomeRange(new_chrom, current_range.start, current_range.end)
            range_ = browser.chrom_lengthes.bound_range(range_)
            browser.goto(range_, who='chromosomes_list')
            browser.refresh()

        self.widgets['chromosomes_list'].observe(chrom_dropdown_val_change, names="value")

        # left_button click
        def left_button_click(b):
            browser.go_left()
            browser.refresh()

        #            browser.preload_imgs('left')
        self.widgets['left_button'].on_click(left_button_click)

        # right_button click
        def right_button_click(b):
            browser.go_right()
            browser.refresh()

        #            browser.preload_imgs('right')
        self.widgets['right_button'].on_click(right_button_click)

        # zoom_in_button click
        def zoom_in_button_click(b):
            browser.zoom_in()
            browser.refresh()

        #            browser.preload_imgs('zoom-in')
        self.widgets['zoom_in_button'].on_click(zoom_in_button_click)

        # zoom_out_button click
        def zoom_out_button_click(b):
            browser.zoom_out()
            browser.refresh()

        #            browser.preload_imgs('zoom-out')
        self.widgets['zoom_out_button'].on_click(zoom_out_button_click)

        # go_button click
        def go_button_click(b):
            range_str = self.widgets['range_textbox'].value.strip("'")
            range_ = GenomeRange(range_str)
            browser.goto(range_, who='go_button')
            browser.refresh()

        self.widgets['go_button'].on_click(go_button_click)

        # range_slider value change
        def range_slider_val_change(change):
            start_old, end_old = change['old']
            length_old = end_old - start_old

            start, end = change['new']
            chrom = browser.current_range.chrom
            if end - start <= 0:
                end = start + length_old
            new_range = GenomeRange(chrom, start, end)
            new_range = browser.chrom_lengthes.bound_range(new_range)
            browser.goto(new_range, who='range_slider')
            browser.refresh()

        self.widgets['range_slider'].observe(range_slider_val_change, names="value")

        # auto_check_box value change
        def auto_check_box_val_change(change):
            if change['new'] == True:
                self.widgets['track_min_val_float_text'].disabled = True
                self.widgets['track_max_val_float_text'].disabled = True
                self.widgets['track_dropdown'].disabled = True
                browser.frame.set_tracks_min_max('auto', 'auto')
            else:
                self.widgets['track_min_val_float_text'].disabled = False
                self.widgets['track_max_val_float_text'].disabled = False
                self.widgets['track_dropdown'].disabled = False
                min_ = self.widgets['track_min_val_float_text'].value
                max_ = self.widgets['track_max_val_float_text'].value
                browser.frame.set_tracks_min_max(min_, max_)
            browser.clear_fig_cache()
            browser.refresh()

        self.widgets['auto_check_box'].observe(auto_check_box_val_change, names="value")

        # track_max_value_float_text value change and
        #   track_min_value_float_text value change
        def track_float_text_val_change(change):
            min_ = self.widgets['track_min_val_float_text'].value
            max_ = self.widgets['track_max_val_float_text'].value
            if self.widgets['track_dropdown'].value == ALL_BW_MARK:
                browser.frame.set_tracks_min_max(min_, max_)
            else:
                for tname in self.selected_tracks:
                    browser.frame.set_tracks_min_max(min_, max_, tname)
            browser.clear_fig_cache()
            browser.refresh()

        self.widgets['track_min_val_float_text'].observe(track_float_text_val_change, names="value")
        self.widgets['track_max_val_float_text'].observe(track_float_text_val_change, names="value")

        def track_dropdown_change(change):
            if change['new'] == ALL_BW_MARK:
                self.selected_tracks = self.__get_tracks_name(browser)
            else:
                self.selected_tracks = [change['new']]

        self.widgets['track_dropdown'].observe(track_dropdown_change, names="value")

    def __get_widgets(self, chromosomes, browser, frame=None):
        if frame is None:
            frame = HTML()
        tracks = self.__get_tracks_name(browser)
        widgets = OrderedDict([
            ("chromosomes_list", Dropdown(options=chromosomes)),
            ("left_button", Button(icon="arrow-left")),
            ("right_button", Button(icon="arrow-right")),
            ("zoom_out_button", Button(icon="search-minus")),
            ("zoom_in_button", Button(icon="search-plus")),
            ("range_textbox", Text(placeholder="genome range like: 'chr1:10000-20000'")),
            ("go_button", Button(description="Go")),

            ("range_slider", IntRangeSlider(continuous_update=False, readout=False, layout=Layout(width='90%'))),
            ("range_min_label", Label("", layout=Layout(width='2%'))),
            ("range_max_label", Label("", layout=Layout(width='20%'))),

            ("auto_check_box", Checkbox(value=True, description="Auto Range",
                                        layout=Layout(width='120px'),
                                        style={'description_width': 'initial'})),
            ("track_min_val_float_text",
             FloatText(value=0.0001, description="Track's min value:", step=0.5, disabled=True,
                       layout=Layout(width='30%'),
                       style={'description_width': 'initial'})),
            ("track_max_val_float_text", FloatText(value=10, description="Track's max value:", step=0.5, disabled=True,
                                                   layout=Layout(width='30%'),
                                                   style={'description_width': 'initial'})),
            ("track_dropdown", Dropdown(options=[ALL_BW_MARK] + tracks,
                                        value=ALL_BW_MARK,
                                        description="Select track",
                                        disabled=True,
                                        )),
            ("frame", frame)
        ])
        return widgets

    def __compose_panel(self, widgets_dict):
        panel = VBox([
            VBox([
                HBox(list(widgets_dict.values())[:7]),
                HBox([
                    widgets_dict["range_min_label"],
                    widgets_dict["range_slider"],
                    widgets_dict["range_max_label"],
                ]),
                HBox([
                    widgets_dict["auto_check_box"],
                    widgets_dict["track_min_val_float_text"],
                    widgets_dict["track_max_val_float_text"],
                    widgets_dict["track_dropdown"],
                ], layout=Layout(justify_content="flex-start")),
            ], layout=Layout(border='solid 2px')),
            widgets_dict["frame"],
        ])
        return panel



import logging
import sys

LOG_LEVEL = logging.WARNING


def get_logger(name, file_=sys.stderr, level=LOG_LEVEL):
    FORMAT = "[%(levelname)s:%(filename)s:%(lineno)s - %(funcName)20s()] %(message)s"
    formatter = logging.Formatter(fmt=FORMAT)
    if isinstance(file_, str):
        handler = logging.FileHandler(file_)
    else:
        handler = logging.StreamHandler(file_)
    handler.setFormatter(formatter)
    log = logging.getLogger(name)
    log.addHandler(handler)
    log.setLevel(level)
    return log



from .base import HistBase, HistData, GenomeRange
from coolbox.utilities.bam import process_bam, coverage_by_samtools


class BAMCov(HistBase):
    """
    Alignment reads coverage track.

    Parameters
    ----------
    file: str
        File path of bam file.

    num_bins: int, optional
        Number of bins to plot hist fig.


    """

    DEFAULT_PROPERTIES = {
        "height": 3,
        "style": HistBase.STYLE_FILL,
        "color": "#6688ff",
        "num_bins": 200
    }

    def __init__(self, file, **kwargs):
        properties = BAMCov.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs,
        })
        super().__init__(**properties)
        self.indexed_bam = process_bam(file)

    def fetch_data(self, gr: GenomeRange, **kwargs) -> HistData:
        bins = self.properties.get("num_bins", 200)
        scores_per_bin = self.fetch_coverage(gr, bins)
        return scores_per_bin

    def fetch_coverage(self, genome_range: GenomeRange, bins=100):
        scores_per_bin = coverage_by_samtools(
            self.indexed_bam,
            str(genome_range),
            bins
        )
        return scores_per_bin



from collections import OrderedDict

from ipywidgets import (Tab, VBox, HTML)

from .base import WidgetsBox
from .navigation import NavigationBar


def compose_track_config_panel(widgets_panel):
    return VBox([])


class FullWidgets(WidgetsBox):

    def __init__(self, browser, *args, **kwargs):
        chromosomes = browser.chrom_lengthes.keys()
        frame_widget = HTML()
        self.navigation_bar = NavigationBar(chromosomes, frame_widget)
        super().__init__(browser, frame_widget, *args, **kwargs)

    def get_widgets_dict(self):
        navigation_bar = self.navigation_bar.widgets
        track_config = OrderedDict([

        ])
        widgets_dict = OrderedDict([
            ("navigation_bar", navigation_bar),
            ("track_config", track_config),
        ])
        return widgets_dict

    def compose_panel(self, widgets_dict):
        # compose navigation_bar
        navigation_bar = self.navigation_bar.panel
        track_config = compose_track_config_panel(widgets_dict['track_config'])
        panel = Tab()
        panel.children = [navigation_bar, track_config]
        panel.set_title(0, "Navigation")
        panel.set_title(1, "Tracks")
        return panel

    def register_events_handler(self):
        self.navigation_bar.register_events_handler(self.browser)

    def refresh_widgets(self, who=None):
        self.navigation_bar.refresh_widgets(self.browser, who)



import numpy as np
import pandas as pd
from matplotlib.patches import Arc, Rectangle, Ellipse
from matplotlib import cm

from coolbox.utilities.genome import GenomeRange, get_logger


log = get_logger(__name__)


class PlotContacts(object):

    def plot_contacts(self, ax, gr: GenomeRange, gr2: GenomeRange, intervals: pd.DataFrame):
        style = self.properties['style']
        columns = intervals.columns
        if style == self.STYLE_ARCS:
            assert len(intervals) == 0 or all(c in columns for c in ['pos1', 'pos2']), \
                "The 'arcs' style's input DataFrame should have columns ['pos1', 'pos2']"
            self.plot_arcs(ax, gr, gr2, intervals)
        elif style == self.STYLE_HICPEAKS:
            assert len(intervals) == 0 or all(c in columns for c in ['start1', 'end1', 'start2', 'end2']), \
                "The 'hicpeaks' style's input DataFrame should have columns:" \
                " ['start1', 'end1', 'start2', 'end2'] or ['pos1', 'pos2']"
            self.plot_hicpeaks(ax, gr, gr2, intervals)
        else:
            raise ValueError("Style should be one of ['arcs', 'peaks']")

    # only support for 1 gr
    def plot_arcs(self, ax, gr: GenomeRange, gr2: GenomeRange, intervals: pd.DataFrame):
        """
        Parameters
        ----------
        ax
        gr
        gr2
        intervals : pd.DataFrame
            Should be with columns ['pos1', 'pos2', 'score']. The `score` column is optional
        """
        properties = self.properties

        def get_height(diameter):
            key = 'diameter_to_height'
            if key in properties:
                try:
                    return eval(properties[key])
                except Exception:
                    pass
            return 0.97 * max_height * diameter / max_diameter

        def get_linewidth(score):
            if ('line_width' in properties) and properties['line_width'] is not None:
                return float(properties['line_width'])
            elif 'score_to_width' in properties:
                try:
                    import numpy
                    import math
                    return eval(properties['score_to_width'])
                except Exception as e:
                    log.warning(str(e))
                    log.warning("Fail to evaluate width according to score")
            return 0.5 * np.sqrt(score)

        def adjust_yaxis(height):
            if properties['orientation'] == 'inverted':
                ax.set_ylim(height, 0.001)
            else:
                ax.set_ylim(-0.001, height)

        def get_color(score, min_score, max_score):
            if properties['cmap'] is not None:
                cmap = cm.get_cmap(properties['cmap'])
                return cmap((max(score, min_score)-min_score)/(max_score-min_score))
            else:
                return properties['color']

        alpha = properties['alpha']
        max_height = properties['height']
        adjust_yaxis(max_height)
        ax.set_xlim(gr.start, gr.end)
        if len(intervals) == 0:
            return

        if properties['cmap'] is not None:
            if properties['vmax'] is not None:
                max_score = properties['vmax']
            elif 'score' in intervals.columns:
                max_score = intervals['score'].max()
            else:
                max_score = 1
        else:
            max_score = 1
        min_score = properties['vmin'] or 0

        max_diameter = (intervals['pos2'] - intervals['pos1']).max()
        for row in intervals.itertuples():
            start, end = row.pos1, row.pos2
            score = row.score if hasattr(row, 'score') else 1
            line_width = get_linewidth(score)
            color = get_color(score, min_score, max_score)
            diameter = (end - start)
            height = 2 * get_height(diameter)
            center = (start + end) / 2
            ax.plot([center], [diameter])
            arc = Arc(
                (center, 0), diameter,
                height, 0, 0, 180,
                color=color,
                alpha=alpha,
                lw=line_width,
                linestyle=properties['line_style']
            )
            if properties['fill'] == 'yes':
                fill = Ellipse((center, 0), diameter, height,
                               alpha=properties['fill_alpha'],
                               color=properties['fill_color'],
                               fill=True)
                ax.add_patch(fill)
            ax.add_patch(arc)

    def plot_hicpeaks(self, ax, gr: GenomeRange, gr2: GenomeRange, intervals: pd.DataFrame):
        """ plot hic peaks uppon a HicMatBase track.

        Parameters
        ----------
        ax
        gr
        gr2
        intervals : pd.DatatFrame
            Should be with columns ['start1', 'end1', 'start2', 'end2']. The `score` column is optional
        """
        import math
        from coolbox.core.track.hicmat import HicMatBase
        # coverage only
        assert 'track' in self.__dict__ and isinstance(self.track, HicMatBase), \
            "No parent HicMatBase Track found. The hicpeaks style " \
            "must be plotted on a HicMatBase Track as a coverage track"
        hicmat_tri_style = (HicMatBase.STYLE_WINDOW, HicMatBase.STYLE_TRIANGULAR)
        hicmat_ma_style = (HicMatBase.STYLE_MATRIX,)

        if len(intervals) == 0:
            return

        # Calculate the position and size of the box from loop's start and end postion.
        def pos_triangular(st1, ed1, st2, ed2):
            m1 = (st1 + ed1) / 2
            m2 = (st2 + ed2) / 2
            x = (m1 + m2) / 2
            y = x - m1
            w = ((ed2 - st2) / 2) / math.cos(math.pi / 4)
            h = ((ed1 - st1) / 2) / math.cos(math.pi / 4)
            return x, y, (w, h)

        def pos_matrix_upper(st1, ed1, st2, ed2):
            x = (st2 + ed2) / 2
            y = (st1 + ed1) / 2
            w = ed2 - st2
            h = ed1 - st1
            return x, y, (w, h)

        def pos_matrix_lower(st1, ed1, st2, ed2):
            x = (st1 + ed1) / 2
            y = (st2 + ed2) / 2
            w = ed1 - st1
            h = ed2 - st2
            return x, y, (w, h)

        hictrack = self.track
        hicmat_style = hictrack.properties['style']
        hicmat_depth_ratio = hictrack.properties['depth_ratio']
        hicmat_depth_ratio = hicmat_depth_ratio if hicmat_depth_ratio != HicMatBase.DEPTH_FULL else 1
        depth_limit = hicmat_depth_ratio * (gr.end - gr.start) / 2

        properties = self.properties
        plot_kwargs = {
            'edgecolor': properties['color'],
            'facecolor': properties['fill_color'],
            'linewidth': properties['line_width'],
            'linestyle': properties['line_style'],
            'fill': True if properties['fill'] == 'yes' else False
        }
        side = properties['side']

        def set_alpha(rec):
            fc = list(rec.get_facecolor())
            fc[-1] = properties['fill_alpha']
            rec.set_facecolor(fc)
            ec = list(rec.get_edgecolor())
            ec[-1] = properties['alpha']
            rec.set_edgecolor(ec)
            return rec

        for loop in intervals.itertuples():
            (st1, ed1, st2, ed2) = loop.start1, loop.end1, loop.start2, loop.end2
            if hicmat_style in hicmat_tri_style:
                x, y, (w, h) = pos_triangular(st1, ed1, st2, ed2)
                if y >= depth_limit:
                    continue
                rec = Rectangle((x, y), w, h, angle=45, **plot_kwargs)
                rec = set_alpha(rec)
                ax.add_patch(rec)

            elif hicmat_style in hicmat_ma_style:
                if side in ('upper', 'both'):
                    # plot upper rectangle
                    x, y, (w, h) = pos_matrix_upper(st1, ed1, st2, ed2)
                    rec = Rectangle((x, y), w, h, **plot_kwargs)
                    rec = set_alpha(rec)
                    ax.add_patch(rec)

                if side in ('lower', 'both'):
                    # plot lower rectangle
                    x, y, (w, h) = pos_matrix_lower(st1, ed1, st2, ed2)
                    rec = Rectangle((x, y), w, h, **plot_kwargs)
                    rec = set_alpha(rec)
                    ax.add_patch(rec)



from setuptools import setup, find_packages
import re


classifiers = [
    "Development Status :: 3 - Alpha",
    "Operating System :: POSIX",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Topic :: Scientific/Engineering :: Visualization",
]


keywords = [
    'genomics',
    'bioinformatics',
    'visualization',
    'Jupyter',
]


def get_version():
    with open("coolbox/__init__.py") as f:
        for line in f.readlines():
            m = re.match("__version__ = '([^']+)'", line)
            if m:
                ver = m.group(1)
                return ver
        raise IOError("Version information can not found.")


def get_long_description():
    return "See https://github.com/GangCaoLab/CoolBox"


def get_install_requires():
    requirements = []
    with open('requirements.txt') as f:
        for line in f:
            requirements.append(line.strip())
    return requirements


setup(
    name='coolbox',
    author='Weize Xu',
    author_email='vet.xwz@gmail.com',
    version=get_version(),
    license='GPLv3',
    description='Jupyter notebook based genomic data visulization toolkit.',
    long_description=get_long_description(),
    keywords=keywords,
    url='https://github.com/GangCaoLab/CoolBox',
    packages=find_packages(),
    scripts=['scripts/coolbox'],
    include_package_data=True,
    zip_safe=False,
    classifiers=classifiers,
    install_requires=get_install_requires(),
    python_requires='>=3.7, <4',
)



from pathlib import Path
from typing import Union

from .base import HicMatBase
from .cool import Cool
from .dothic import DotHiC


def HiCMat(file_or_hicmat: Union[str, HicMatBase], *args, **kwargs) -> HicMatBase:
    """
    Compose DotHic or Cool track automatically based on tpye of file extension (.cool, .mcool, .hic)
    """
    if isinstance(file_or_hicmat, HicMatBase):
        return file_or_hicmat
    elif not Path(file_or_hicmat).is_file():
        raise ValueError("The file path does not exist.")

    p = file_or_hicmat
    if p.endswith(".hic"):
        return DotHiC(file_or_hicmat, *args, **kwargs)
    elif p.endswith((".cool", ".mcool")):
        return Cool(file_or_hicmat, *args, **kwargs)
    else:
        raise NotImplementedError(f"File type of {p} not supported for HicMat. "
                                  f"The file type should be one of .cool/.mcool/.hic")



import numpy as np
import pandas as pd
import statsmodels.stats.multitest as smm
from scipy.ndimage import gaussian_filter
from scipy.stats import norm

from coolbox.utilities.genome import GenomeRange
from coolbox.core.track.hicmat.base import HicMatBase
from coolbox.core.track.hicmat.hicmat import HiCMat


class Selfish(HicMatBase):
    """
    Differential chromatin interaction.

    Parameters
    ----------
    hic1 : coolbox.api.track.Cool
        First HiC Track or hic file path(.cool, .mcool, .hic).

    hic2 : coolbox.api.track.Cool
        Second HiC Track or hic file path(.cool, .mcool, .hic).

    hic_args : dict, optional
        Argument to create Hi-C instance, only in use
        when first or second argument is a path.

    sigma0 : float, optional
        Initial sigma, parameter of SELFISH method. Default: 1.6

    s : int, optional
        Iteration count parameter of SELFISH method. Default: 10

    References
    ----------
    Abbas Roayaei Ardakany, Ferhat Ay, Stefano Lonardi,
    Selfish: discovery of differential chromatin interactions via a self-similarity measure,
    Bioinformatics, Volume 35, Issue 14, July 2019, Pages i145–i153,
    https://doi.org/10.1093/bioinformatics/btz362

    """

    DEFAULT_ARGS_HIC = {
        "transform": "no",
        "normalize": "zscore",
        "gaussian_sigma": "no",
        "process_func": "no",
    }

    DEFAULT_PROPERTIES = {
        "style": HicMatBase.STYLE_WINDOW,
        "cmap": "RdPu_r",
        "transform": False,
        "s": 10,
        "sigma0": 1.6,
        "norm": 'log',
        "args_hic": DEFAULT_ARGS_HIC,
    }

    def __init__(self, hic1, hic2, **kwargs):
        # TODO duplicate code with HiCDiff
        properties = Selfish.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(**properties)

        # TODO how to set hic_args in cli mode ?
        if isinstance(hic1, str):
            hic1 = HiCMat(hic1)
        if isinstance(hic2, str):
            hic2 = HiCMat(hic2)
        self.properties.update({
            'hic1': hic1,
            'hic2': hic2,
        })

        self.mat1 = None
        self.mat2 = None

    def fetch_data(self, gr: GenomeRange, **kwargs) -> np.ndarray:
        # TODO find a better way
        # For cases when updating of Frame's depth_ratio leads to un equal matrix sizes for different tracks).
        args_hic = self.properties.copy()
        args_hic.update(self.properties.get('args_hic', {}))
        for hic in ('hic1', 'hic2'):
            self.properties[hic].properties.update(args_hic)

        hic1: HicMatBase = self.properties['hic1']
        hic2: HicMatBase = self.properties['hic2']
        # must set to avoid re-change GenomeRange in window style
        kwargs['gr_updated'] = True
        self.mat1 = mat1 = hic1.fetch_plot_data(gr, **kwargs)
        self.mat2 = mat2 = hic2.fetch_plot_data(gr, **kwargs)
        # what does this used for ?
        # self.zero_indices = (hic1.zero_indices | hic1.nan_indices) | (hic2.zero_indices | hic2.nan_indices)
        pvals = self.diff_matrix(mat1, mat2)

        return pvals

    def fetch_pixels(self, gr: GenomeRange, **kwargs) -> pd.DataFrame:
        pvals = self.fetch_data(gr, **kwargs)
        mat1 = self.mat1
        mat2 = self.mat2
        binsize = self.properties['hic1'].fetched_binsize
        idx_y, idx_x = np.where(pvals <= kwargs.get('threshold', 1e-4))
        ix_up = idx_y <= idx_x
        idx_y, idx_x = idx_y[ix_up], idx_x[ix_up]
        start = (gr.start // binsize) * binsize
        start1 = start + idx_y * binsize
        end1 = start1 + binsize
        start2 = start + idx_x * binsize
        end2 = start2 + binsize
        df = pd.DataFrame({
            'chrom': gr.chrom,
            'start1': start1,
            'end1': end1,
            'start2': start2,
            'end2': end2,
            'value1': mat1[idx_y, idx_x],
            'value2': mat2[idx_y, idx_x],
            'qvalue': pvals[idx_y, idx_x],
        })
        return df

    def _get_scales(self):
        sigma0 = self.properties['sigma0']
        s = self.properties['s']
        return [(sigma0 * (2 ** (i / s))) for i in range(1, s + 3)]

    def diff_matrix(self, mat1: np.ndarray, mat2: np.ndarray) -> np.ndarray:
        diff = mat2 - mat1
        scales = self._get_scales()
        np.nan_to_num(diff, copy=False, posinf=0, neginf=0, nan=0)
        d_pre = gaussian_filter(diff, scales[0])
        final_p = np.ones(mat1.shape, dtype=mat1.dtype)
        for scale in scales[1:]:
            d_post = gaussian_filter(diff, scale)
            d_diff = d_post - d_pre
            params = norm.fit(d_diff)
            pvals = norm.cdf(d_diff, loc=params[0], scale=params[1])
            np.nan_to_num(pvals, copy=False, posinf=1, neginf=1, nan=1)
            pvals[pvals > 0.5] = 1 - pvals[pvals > 0.5]
            pvals *= 2
            lt_idx = pvals < final_p
            final_p[lt_idx] = pvals[lt_idx]
            d_pre = d_post.copy()
        _, out_p = smm.multipletests(final_p.ravel(), method='fdr_bh')[:2]
        out_p = out_p.reshape(*diff.shape)
        thr = 12
        out_p[out_p == 0] = 10 ** -1 * thr
        return out_p



def cm2inch(*tupl):
    """
    convert length unit from cm to inch.

    >>> cm2inch(10)
    3.937007874015748
    >>> cm2inch(2, 2)
    (0.7874015748031495, 0.7874015748031495)
    >>> cm2inch((1, 5))
    (0.39370078740157477, 1.968503937007874)
    """
    inch = 2.54
    if isinstance(tupl[0], tuple):
        return tuple(i / inch for i in tupl[0])
    else:
        if len(tupl) != 1:
            return tuple(i / inch for i in tupl)
        else:
            return tupl[0] / inch


def inch2cm(*tupl):
    """
    convert length unit from inch to cm.

    """
    cm = 2.54
    if isinstance(tupl[0], tuple):
        return tuple(i * cm for i in tupl[0])
    else:
        if len(tupl) != 1:
            return tuple(i * cm for i in tupl)
        else:
            return tupl[0] * cm


def rgb2hex(r, g, b):
    """
    Convert rgb color to hex format.

    >>> rgb2hex(129, 154, 70)
    '#819a46'
    >>> rgb2hex(-10, 256, -1)
    Traceback (most recent call last):
    ...
    AssertionError: (r, g, b) value must within range 0 ~ 255.
    """
    assert (0, 0, 0) <= (r, g, b) <= (255, 255, 255), \
        "(r, g, b) value must within range 0 ~ 255."
    hex = "#{:02x}{:02x}{:02x}".format(r, g, b)
    return hex


def hex2rgb(color_hex):
    """
    Convert hex color code to rgb tuple.

    >>> hex2rgb('#819a46')
    (129, 154, 70)
    """
    code = color_hex[1:]
    r = int(code[0:2], 16)
    g = int(code[2:4], 16)
    b = int(code[4:6], 16)
    return r, g, b


def shiftedColorMap(cmap, start=0, midpoint=0.5, stop=1.0, name='shiftedcmap'):
    '''
    This function is from: https://stackoverflow.com/a/20528097/8500469

    Function to offset the "center" of a colormap. Useful for
    data with a negative min and positive max and you want the
    middle of the colormap's dynamic range to be at zero

    Input
    -----
      cmap : The matplotlib colormap to be altered
      start : Offset from lowest point in the colormap's range.
          Defaults to 0.0 (no lower ofset). Should be between
          0.0 and `midpoint`.
      midpoint : The new center of the colormap. Defaults to
          0.5 (no shift). Should be between 0.0 and 1.0. In
          general, this should be  1 - vmax/(vmax + abs(vmin))
          For example if your data range from -15.0 to +5.0 and
          you want the center of the colormap at 0.0, `midpoint`
          should be set to  1 - 5/(5 + 15)) or 0.75
      stop : Offset from highets point in the colormap's range.
          Defaults to 1.0 (no upper ofset). Should be between
          `midpoint` and 1.0.

    '''
    import numpy as np
    import matplotlib
    import matplotlib.pyplot as plt

    cdict = {
        'red': [],
        'green': [],
        'blue': [],
        'alpha': []
    }

    # regular index to compute the colors
    reg_index = np.linspace(start, stop, 257)

    # shifted index to match the data
    shift_index = np.hstack([
        np.linspace(0.0, midpoint, 128, endpoint=False),
        np.linspace(midpoint, 1.0, 129, endpoint=True)
    ])

    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))

    newcmap = matplotlib.colors.LinearSegmentedColormap(name, cdict)
    plt.register_cmap(cmap=newcmap)

    return newcmap


def get_size(obj, seen=None):
    """
    Recursively finds size of objects

    From:
        https://stackoverflow.com/a/40880923/8500469
    """
    import sys
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size


def fig2bytes(fig, encode='svg', dpi=None):
    """
    Convert matplotlib.figure.Figure/IPython.display.SVG
    object to image bytes.
    """
    from IPython.display import SVG
    if isinstance(fig, SVG):
        img_bytes = fig.data.encode('utf-8')
    else:
        import io
        buf = io.BytesIO()
        fig.savefig(buf, format=encode, dpi=None)
        buf.seek(0)
        img_bytes = buf.read()
        buf.close()
    return img_bytes



import numpy as np
import pandas as pd

from coolbox.utilities import (
    split_genome_range, change_chrom_names,
    GenomeRange, get_logger, to_gr
)
from .base import HistBase

log = get_logger(__name__)


class BigWig(HistBase):
    """
    BigWig track

    Parameters
    -----------
    file : str
        File path of bigwig file.

    num_bins : int, optional
        Number of bins to plot the hist in current range, default 700.


    """

    DEFAULT_PROPERTIES = {
        "color": "#dfccde",
        "style": HistBase.STYLE_FILL,
        "num_bins": 700,
        "threshold": "inf"
    }

    def __init__(self, file, **kwargs):
        properties = BigWig.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs
        })
        super().__init__(**properties)
        import bbi
        self.bw = bbi.open(self.properties['file'])

    def fetch_plot_data(self, gr: GenomeRange, **kwargs):
        num_bins = self.get_num_bins()
        self.check_chrom_name(gr)
        scores_per_bin = self.fetch_scores(gr, num_bins)
        return scores_per_bin

    def fetch_data(self, gr: GenomeRange, **kwargs):
        """
        Parameters
        ----------
        gr : GenomeRange

        Return
        ------
        intervals : pandas.core.frame.DataFrame
            BigWig interval table.
        """
        chrom, start, end = split_genome_range(gr)
        if chrom not in self.bw.chromsizes:
            chrom = change_chrom_names(chrom)

        intervals = self.bw.fetch_intervals(chrom, start, end)
        columns = list(intervals.columns)
        if 'value' in columns:
            columns[columns.index('value')] = 'score'
        intervals.columns = columns

        return intervals

    def get_num_bins(self, default_num=700):
        num_bins = default_num
        if 'number_of_bins' in self.properties:
            try:
                num_bins = int(self.properties['number_of_bins'])
            except TypeError:
                num_bins = default_num
                log.warning("'number_of_bins' value: {} for bigwig file {} "
                            "is not valid. Using default value (700)".format(self.properties['number_of_bins'],
                                                                             self.properties['file']))
        return num_bins

    def fetch_scores(self, genome_range, num_bins, max_try_nums=5):
        """Fetch bins scores within input chromosome range.
        """
        scores_per_bin = np.zeros(num_bins)
        gr = to_gr(genome_range)
        if gr.chrom not in self.bw.chromsizes:
            gr.change_chrom_names()
        try:
            scores_per_bin = self.bw.fetch(gr.chrom, gr.start, gr.end, num_bins).astype(float)
        except Exception as e:
            log.warning(f"error found while reading bigwig scores: {e}")
            pass
        return scores_per_bin

    def check_chrom_name(self, genome_range):
        if genome_range.chrom not in self.bw.chromsizes:
            genome_range.change_chrom_names()

        if genome_range.chrom not in self.bw.chromsizes:
            log.warning("Can not read region {} from bigwig file:\n\n"
                        "{}\n\nPlease check that the chromosome name is part of the bigwig file "
                        "and that the region is valid".format(str(genome_range), self.properties['file']))


class ABCompartment(BigWig):
    """
    A/B Compartment BigWig track.
    """

    DEFAULT_PROPERTIES = {
        "color": "#66ccff",
        "threshold_color": "#ff9c9c",
        "threshold": 0,
    }

    def __init__(self, file, **kwargs):
        properties = ABCompartment.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(file, **properties)



import pandas as pd

from .base import ArcsBase
from .fetch import FetchParix
from coolbox.utilities.bed import process_pairs
from coolbox.utilities.genome import GenomeRange


class Pairs(ArcsBase, FetchParix):
    """
    Arcs track from .pairs file.

    Parameters
    ----------
    file: str
        Path of .pairs file

    """
    DEFAULT_PROPERTIES = {
        'color': '#dc9732'
    }
    FIELDS = ["name", "chr1", "pos1", "chr2", "pos2", "strand1", "strand2"]

    def __init__(self, file, **kwargs):
        properties = Pairs.DEFAULT_PROPERTIES.copy()
        properties.update({
            'file': file,
            **kwargs
        })
        super().__init__(**properties)
        self.bgz_file = process_pairs(file)

    def fetch_data(self, gr: GenomeRange, **kwargs):
        # filter peaks manually in peaks style
        df = self.fetch_intervals(self.bgz_file, gr, kwargs.get('gr2'))
        # TODO the returned df has no named columns, may cause error
        if len(df) == 0:
            return df

        columns = list(df.columns)
        for i, col in enumerate(self.FIELDS):
            if i >= len(columns):
                break
            columns[i] = col
        df.columns = columns
        for col in ['pos1', 'pos2']:
            df[col] = df[col].astype(int)
        return df



from .highlights import HighLights, HighLightsFromFile
from .vlines import Vlines, VlinesFromFile
from .hlines import HLines

from .base import track_to_coverage
from ..track.bed import TAD
from ..track.hist import Hist, BigWig, BedGraph
from ..track.arcs import Arcs, Pairs, BEDPE, HiCPeaks

TADCoverage = track_to_coverage(TAD)

ArcsCoverage = track_to_coverage(Arcs)
PairsCoverage = track_to_coverage(Pairs)
BEDPECoverage = track_to_coverage(BEDPE)
HiCPeaksCoverage = track_to_coverage(HiCPeaks)

HistCoverage = track_to_coverage(Hist)
BigWigCoverage = track_to_coverage(BigWig)
BedGraphCoverage = track_to_coverage(BedGraph)



import os
import os.path as osp
import uuid


def opener(filename):
    """
    Determines if a file is compressed or not

    >>> import gzip
    >>> msg = "hello blablabla"
    >>> tmp_f_raw  = open("/tmp/test_opener.txt", 'w')
    >>> tmp_f_raw.write(msg)
    15
    >>> tmp_f_raw.close()
    >>> tmp_f_gzip = gzip.open('/tmp/test_opener.txt.gz', 'wb')
    >>> tmp_f_gzip.write(to_bytes(msg))
    15
    >>> tmp_f_gzip.close()

    >>> test_raw = opener(tmp_f_raw.name)
    >>> type(test_raw)
    <class '_io.BufferedReader'>
    >>> test_gzip = opener(tmp_f_gzip.name)
    >>> type(test_gzip)
    <class 'gzip.GzipFile'>

    >>> test_raw.close()
    >>> test_gzip.close()

    >>> import os
    >>> os.remove(test_raw.name)
    >>> os.remove(test_gzip.name)

    """
    import gzip
    f = open(filename, 'rb')
    if f.read(2) == b'\x1f\x8b':
        f.seek(0)
        return gzip.GzipFile(fileobj=f)
    else:
        f.seek(0)
        return f


def to_string(s):
    """
    Convert bytes, bytes list to string, string list.

    >>> to_string("hello")
    'hello'
    >>> to_string(b"hello")
    'hello'
    >>> to_string([b'hello', b'world'])
    ['hello', 'world']
    """
    if isinstance(s, str):
        return s
    if isinstance(s, bytes):
        return s.decode('ascii')
    if isinstance(s, list):
        return [to_string(x) for x in s]
    return s


def to_bytes(s):
    """
    Like toString, to bytes.

    >>> to_bytes('hello')
    b'hello'
    >>> to_bytes(['hello', 'world'])
    [b'hello', b'world']
    """
    if isinstance(s, bytes):
        return s
    if isinstance(s, str):
        return bytes(s, 'ascii')
    if isinstance(s, list):
        return [to_bytes(x) for x in s]
    return s


def get_tmp_dir(name=".coolbox"):
    """Return the path to coolbox temporary directory.
    If the file not exists, will make it.
    """
    path = osp.join(osp.realpath(os.curdir), name)
    if not osp.exists(path):
        os.mkdir(path)
    return path


def get_uniq_tmp_file(prefix="", suffix='.svg', dirname=".coolbox"):
    """Return the path to a unique(not exists) temporary file."""
    tmp_dir = get_tmp_dir(dirname)
    while True:
        short_uuid = str(uuid.uuid1()).split('-')[0]
        file = prefix + short_uuid + suffix
        path = osp.join(tmp_dir, file)
        if not osp.exists(path):
            break
    return path



from copy import copy
from collections import OrderedDict

import matplotlib
import matplotlib.pyplot as plt
import mpl_toolkits.axisartist as axisartist

from coolbox.utilities import (
    cm2inch,
    GenomeRange,
    get_logger,
)
from .base import FrameBase

log = get_logger(__name__)

plt.rcParams['svg.fonttype'] = 'none'


class Frame(FrameBase):
    """
    Frame for arrange and group plots.

    Parameters
    ----------
    genome_range : {GenomeRange, str}, optional
        The init genome range.

    width : {int, float}, optional
        Width of frame, the unit is in 'cm', 1cm = 2.54inch,
        default Frame.DEFAULT_WIDTH.

    width_ratios : tuple, optional
        Width ratios of track and track title,
        like `(0.9, 0.1)`. default Frame.DEFAULT_WIDTH_RATIOS.

    margins : dict, optional
        Margins of frame, default Frame.DEFAULT_MARGINS.

    title : str, optional
        The title of this frame, default ''.

    Attributes
    ----------
    properties : dict
        Frame properties dict.

    tracks : OrderedDict
        Container of all tracks.

    current_range : GenomeRange, optional
        Current frame range.

    Examples
    --------
    >>> frame_1 = Frame()
    >>> frame_2 = Frame(gr="chr1:1000-2000")
    >>> str(frame_2.current_range)
    'chr1:1000-2000'
    >>> frame_3 = Frame(gr=GenomeRange("chr1", 1000, 2000))
    >>> str(frame_3.current_range)
    'chr1:1000-2000'
    """

    DEFAULT_WIDTH = 40
    DEFAULT_WIDTH_RATIOS = (0.01, 0.93, 0.06)
    DEFAULT_MARGINS = {'left': 0.04, 'right': 0.92, 'bottom': 0, 'top': 1}

    def __init__(self, *args, **kwargs):

        tracks = OrderedDict()
        self._tracks = tracks
        properties_dict = {
            "width": Frame.DEFAULT_WIDTH,
            "width_ratios": Frame.DEFAULT_WIDTH_RATIOS,
            "margins": Frame.DEFAULT_MARGINS,
            "title": ""
        }
        properties_dict.update(kwargs)
        super().__init__(properties_dict, *args, **kwargs)

    @property
    def tracks(self):
        return self._tracks

    def show(self):
        """
        Display self's elements on screen.
        """
        if self.current_range is None:
            raise ValueError("Frame range is uninitialized, "
                             "please `Frame.goto` method to initialize it.")

        return self.plot(self.current_range.chrom,
                         self.current_range.start,
                         self.current_range.end)

    def add_track(self, track, pos='tail'):
        """
        Add `Track` object to self.

        Parameters
        ----------
        track : Track
            The track need to be added to self.tracks

        pos : {'tail', 'head'}
            Add track to tail or head. default 'tail'

        Examples
        --------
        >>> from coolbox.core.track import XAxis, BigWig
        >>> frame = Frame()
        >>> frame.add_track(XAxis())
        >>> frame.add_track(BigWig("tests/test_data/bigwig_chrx_2e6_5e6.bw"))
        >>> len(frame.tracks)
        2
        """
        from ..track.base import Track

        assert isinstance(track, Track), "track must a Track object."
        if pos == 'tail':
            self.tracks.update({track.name: track})
        else:
            self.tracks.update({track.name: track})
            self.tracks.move_to_end(track.name, last=False)

    def get_tracks_height(self, default_height=3):
        """
        Get heights of all tracks.

        Return
        ------
        heights : list of float
            heights of all tracks.
        """
        heights = []
        for track in self.tracks.values():
            if hasattr(track, 'get_track_height'):
                frame_width = self.properties['width'] * self.properties['width_ratios'][1]
                height = track.get_track_height(frame_width)
                heights.append(height)
            elif 'height' in track.properties:
                heights.append(track.properties['height'])
            else:
                heights.append(default_height)
        return heights

    def plot(self, *args):
        """
        Plot all tracks.

        >>> from coolbox.api import *
        >>> frame = XAxis() + XAxis()
        >>> frame.plot("chr1", 100000, 200000)
        >>> frame.plot("chr1:100000-200000")
        """
        if len(args) == 3:
            gr = GenomeRange(*args[:3])
            gr2 = None
        elif len(args) == 1:
            gr = GenomeRange(args[0])
            gr2 = None
        elif len(args) == 6:
            gr = GenomeRange(args[:3])
            gr2 = GenomeRange(args[3:])
        elif len(args) == 2:
            gr = GenomeRange(args[0])
            gr2 = GenomeRange(args[1])
        elif self.current_range:
            gr = self.current_range
            gr2 = None
        else:
            raise ValueError("Please specify a genomic range in uscs format. For example: 'chr1:100000-200000'")
        # cache for the previous GenomeRange
        self.goto(gr, gr2)

        tracks_height = self.get_tracks_height()
        self.properties['height'] = sum(tracks_height)

        fig = plt.figure(figsize=cm2inch(self.properties['width'],
                                         self.properties['height']))
        if 'title' in self.properties:
            fig.suptitle(self.properties['title'])

        grids = matplotlib.gridspec.GridSpec(
            len(tracks_height), 3,
            height_ratios=tracks_height,
            width_ratios=self.properties['width_ratios'],
            wspace=0.01)

        axis_list = []
        for idx, track in enumerate(self.tracks.values()):
            y_ax = plt.subplot(grids[idx, 0])
            y_ax.set_axis_off()

            ax = axisartist.Subplot(fig, grids[idx, 1])
            fig.add_subplot(ax)
            ax.axis[:].set_visible(False)
            ax.patch.set_visible(False)

            label_ax = plt.subplot(grids[idx, 2])
            label_ax.set_axis_off()

            track.label_ax = label_ax
            track.y_ax = y_ax
            try:
                # Attention, copy is necessary, otherwise GenomeRange may change due to call of gr.change_chrom_names
                track.plot(ax, copy(gr), gr2=copy(gr2))

            except Exception as e:
                import sys, os
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = exc_tb.tb_frame.f_code.co_filename
                log.error("Error occured when plot track:\n"
                          "\ttrack name: {}\n\ttrack type:{}\n"
                          "\tError: {} {}\n"
                          "\toccurred in \"{}\", line {}".format(
                    track.name, type(track),
                    type(e), str(e), fname, exc_tb.tb_lineno)
                )
                log.exception(e)
            track.plot_coverages(ax, gr, gr2)

            axis_list.append(ax)

        margins = self.properties['margins']
        fig.subplots_adjust(wspace=0, hspace=0.0,
                            left=margins['left'],
                            right=margins['right'],
                            bottom=margins['bottom'],
                            top=margins['top'])

        plt.close()

        return fig


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from .bed import BED
from .base import BedBase


class TAD(BED):
    """
    Tad tack from bed file
    """

    DEFAULT_PROPERTIES = {
        'style': BedBase.STYLE_TAD,
        'alpha': 0.3,
    }

    def __init__(self, file, **kwargs):
        properties = TAD.DEFAULT_PROPERTIES.copy()
        properties.update(kwargs)
        super().__init__(file, **properties)



from coolbox.utilities.logtools import get_logger
from coolbox.utilities.genome import to_gr

log = get_logger(__name__)


class StrawWrap(object):
    """
    A wrap for straw Python API, for read .hic file.

    Parameters
    ----------
    path : str
        Path to the '.hic' file

    normalization : {False, 'VC', 'VC_SQRT', 'KR'}
        Method for the matrix normalization.
        default 'KR'

    binsize : {'auto', int}
        resolution of the data. for example 5000.
        'auto' for calculate resolution automatically.
        default 'auto'

    """

    def __init__(self, path, normalization='KR', binsize='auto'):
        self.hic_file = path
        if normalization == 'no' or normalization is False:
            normalization = 'NONE'
        elif normalization is True:
            normalization = 'KR'
        self.normalization = normalization
        self.binsize = binsize
        self.chromosomes, self.resolutions, self.masterindex, self.genome, self.metadata = self.__info()
        self.fetched_binsize = None

    def fetch(self, genome_range1, genome_range2=None):
        """
        Return
        ------
        matrix : numpy.ndarray
        """
        from coolbox.utilities.genome import GenomeRange

        if genome_range2 is None:
            genome_range2 = genome_range1
        genome_range1 = to_gr(genome_range1)
        genome_range2 = to_gr(genome_range2)

        if isinstance(genome_range1, str):
            genome_range1 = GenomeRange(genome_range1)
        if isinstance(genome_range2, str):
            genome_range2 = GenomeRange(genome_range2)

        if genome_range1.chrom.startswith("chr"):
            genome_range1.change_chrom_names()
        if genome_range2.chrom.startswith("chr"):
            genome_range2.change_chrom_names()

        binsize = self.infer_binsize(genome_range1)
        self.fetched_binsize = binsize  # expose fetched binsize

        straw_list = self.__fetch_straw_list(genome_range1, genome_range2, binsize)
        matrix = self.__list_to_matrix(straw_list, genome_range1, genome_range2, binsize)
        return matrix

    def infer_binsize(self, genome_range):
        from .tools import infer_resolution
        if self.binsize == 'auto':
            binsize = infer_resolution(genome_range, self.resolutions)
        else:
            binsize = self.binsize
        return binsize

    def __fetch_straw_list(self, genome_range1, genome_range2, binsize):
        from coolbox.utilities.hic.straw import straw
        chr1loc = str(genome_range1).replace('-', ':')
        chr2loc = str(genome_range2).replace('-', ':')
        slist = []
        try:
            slist = straw(self.normalization, self.hic_file, chr1loc, chr2loc, 'BP', binsize)
        except Exception as e:
            log.warning("Error occurred when reading the dothic file with straw:")
            log.warning(str(e))
            if self.normalization != "NONE":
                log.warning("Try to read unbalanced matrix.")
                self.normalization = "NONE"
                try:
                    slist = straw(self.normalization, self.hic_file, chr1loc, chr2loc, 'BP', binsize)
                except Exception as e:
                    log.warning("Failed.")
                    log.warning(str(e))
                log.warning("Unbalanced matrix is readed.")
        return slist

    def fetch_pixels(self, genome_range1, genome_range2=None):
        from pandas import DataFrame
        if genome_range2 is None:
            genome_range2 = genome_range1
        if genome_range1.chrom.startswith("chr"):
            genome_range1.change_chrom_names()
        if genome_range2.chrom.startswith("chr"):
            genome_range2.change_chrom_names()
        binsize = self.infer_binsize(genome_range1)
        slist = self.__fetch_straw_list(genome_range1, genome_range2, binsize)
        pixels = DataFrame(slist)
        pixels = pixels.T
        pixels.columns = ['start1', 'start2', 'value']
        pixels['start1'] = pixels['start1'].astype('int32')
        pixels['start2'] = pixels['start2'].astype('int32')
        pixels['end1'] = pixels['start1'] + binsize
        pixels['end2'] = pixels['start2'] + binsize
        pixels['chrom1'] = genome_range1.chrom
        pixels['chrom2'] = genome_range2.chrom
        pixels = pixels[['chrom1', 'start1', 'end1', 'chrom2', 'start2', 'end2', 'value']]
        return pixels

    def __list_to_matrix(self, straw_list, genome_range1, genome_range2, binsize):
        import numpy as np
        binlen1 = (genome_range1.length // binsize) + 1
        binlen2 = (genome_range2.length // binsize) + 1
        mat = np.zeros((binlen1, binlen2), dtype=np.float64)
        is_cis = (genome_range1 == genome_range2)
        for rec in zip(*straw_list):
            loc1, loc2, c = rec[0], rec[1], rec[2]
            bin1id = (loc1 - genome_range1.start) // binsize
            bin2id = (loc2 - genome_range2.start) // binsize
            if is_cis:
                mat[bin1id, bin2id] = c
                mat[bin2id, bin1id] = c
            else:
                if (0 <= bin1id < mat.shape[0]) and (0 <= bin2id < mat.shape[1]):
                    mat[bin1id, bin2id] = c
        return mat

    def __info(self):
        """
        from hic2cool code:
            https://github.com/4dn-dcic/hic2cool/blob/master/hic2cool/hic2cool_utils.py#L73

        Takes in a .hic file and returns a dictionary containing information about
        the chromosome. Keys are chromosome index numbers (0 through # of chroms
        contained in file) and values are [chr idx (int), chr name (str), chrom
        length (str)]. Returns the masterindex used by the file as well as the open
        file object.
        """
        import sys
        import struct
        from .straw import readcstr

        req = open(self.hic_file, 'rb')

        chrs = {}
        resolutions = []
        magic_string = struct.unpack(b'<3s', req.read(3))[0]
        req.read(1)
        if (magic_string != b"HIC"):
            print('This does not appear to be a HiC file; '
                  'magic string is incorrect')
            sys.exit()
        global version
        version = struct.unpack(b'<i', req.read(4))[0]
        masterindex = struct.unpack(b'<q', req.read(8))[0]
        genome = b""
        c = req.read(1)
        while (c != b'\0'):
            genome += c
            c = req.read(1)
        genome = genome.decode('ascii')
        # metadata extraction
        metadata = {}
        nattributes = struct.unpack(b'<i', req.read(4))[0]
        for x in range(nattributes):
            key = readcstr(req)
            value = readcstr(req)
            metadata[key] = value
        nChrs = struct.unpack(b'<i', req.read(4))[0]
        for i in range(0, nChrs):
            name = readcstr(req)
            length = struct.unpack(b'<i', req.read(4))[0]
            if name and length:
                chrs[i] = [i, name, length]
        nBpRes = struct.unpack(b'<i', req.read(4))[0]
        # find bp delimited resolutions supported by the hic file
        for x in range(0, nBpRes):
            res = struct.unpack(b'<i', req.read(4))[0]
            resolutions.append(res)
        return chrs, resolutions, masterindex, genome, metadata


class CoolerWrap(object):
    """
    wrap for cooler file,
    deal with multi resolution.

    Parameters
    ----------
    path : str
        Path to the cooler file.

    binsize : {'auto', int}
        resolution of the data. for example 5000.
        'auto' for calculate resolution automatically.
        default 'auto'

    balance : bool
        Balance the matrix or not.
        default True
    """

    def __init__(self, path, binsize='auto', balance=True):
        import cooler
        from .tools import is_multi_cool, get_cooler_resolutions
        self.path = path

        self.is_multi = is_multi_cool(path)
        self.resolutions = get_cooler_resolutions(path, self.is_multi)
        if self.is_multi:
            self.coolers = self.__load_multi_coolers(path)
        else:
            self.cool = cooler.Cooler(path)

        self.binsize = binsize
        self.balance = balance

    def __load_multi_coolers(self, path):
        from collections import OrderedDict
        import cooler

        def cooler_reso(resolution):
            from h5py import File
            with File(path, 'r') as f:
                if "resolutions" in f:
                    c = cooler.Cooler(path + "::/resolutions/{}".format(resolution))
                else:
                    for grp_name in f:
                        grp = f[grp_name]
                        if str(grp.attrs['bin-size']) == str(resolution):
                            c = cooler.Cooler(path + "::{}".format(grp_name))
                            break
            return c

        coolers = OrderedDict([(reso, cooler_reso(reso)) for reso in self.resolutions])
        return coolers

    def get_cool(self, genome_range):
        binsize = self.infer_binsize(genome_range)
        if self.is_multi:
            cool = self.coolers[binsize]
        else:
            cool = self.cool
        self.fetched_binsize = binsize  # expose fetched binsize
        return cool

    def infer_binsize(self, genome_range):
        from .tools import infer_resolution
        genome_range = to_gr(genome_range)
        if self.is_multi:
            resolutions = [k for k in self.coolers.keys()]
            if self.binsize == 'auto':
                binsize = infer_resolution(genome_range, resolutions)
            else:
                assert self.binsize in resolutions, \
                    "Multi-Cooler file not contain the resolution {}.".format(self.binsize)
                binsize = int(self.binsize)
        else:
            binsize = self.cool.binsize
        return binsize

    def fetch(self, genome_range1, genome_range2=None):
        # TODO what if genome_ranges are invalid
        if genome_range2 is None:
            genome_range2 = genome_range1

        genome_range1 = to_gr(genome_range1)
        genome_range2 = to_gr(genome_range2)

        cool = self.get_cool(genome_range1)

        if genome_range1.chrom not in cool.chromnames:
            genome_range1.change_chrom_names()
        if genome_range2.chrom not in cool.chromnames:
            genome_range2.change_chrom_names()

        try:
            mat = cool.matrix(balance=self.balance).fetch(str(genome_range1), str(genome_range2))
        except ValueError as e:
            log.warning(str(e))
            log.warning("Data is not balanced, force to use unbalanced matrix.")
            mat = cool.matrix(balance=False).fetch(str(genome_range1), str(genome_range2))

        return mat

    def fetch_pixels(self, genome_range1, genome_range2=None, join=True):
        cool = self.get_cool(genome_range1)

        if genome_range2 is None:
            genome_range2 = genome_range1

        if genome_range1.chrom not in cool.chromnames:
            genome_range1.change_chrom_names()
        if genome_range2.chrom not in cool.chromnames:
            genome_range2.change_chrom_names()

        mat = cool.matrix(as_pixels=True, balance=self.balance, join=join)
        pixels = mat.fetch(str(genome_range1), str(genome_range2))

        return pixels



import typing as T
import subprocess as subp
from functools import partial
from os import path as osp

from intervaltree import Interval, IntervalTree
import collections
import types

from .filetool import opener, to_string
from .logtools import get_logger
from .genome import GenomeRange

log = get_logger(__name__)


def file_to_intervaltree(file_name):
    """
    converts a BED like file into a bx python interval tree

    Parameters
    ----------
    file_name : str
        Path to file.

    Return
    ------
    interval_tree : dict
        Interval tree dictionary. They key is the chromosome/contig name and the
        value is an IntervalTree. Each of the intervals have as 'value' the fields[3:] if any.
    """
    import sys
    import numpy as np

    # iterate over a BED like file
    # saving the data into an interval tree
    # for quick retrieval
    file_h = opener(file_name)
    line_number = 0
    valid_intervals = 0
    prev_chrom = None
    prev_start = -1
    prev_line = None
    interval_tree = {}
    min_value = np.inf
    max_value = -np.inf

    for line in file_h.readlines():
        line_number += 1
        line = to_string(line)
        if line.startswith('browser') or line.startswith('track') or line.startswith('#'):
            continue
        fields = line.strip().split('\t')
        try:
            chrom, start, end = fields[0:3]
        except Exception as detail:
            msg = "Error reading line: {}\nError message: {}".format(line_number, detail)
            sys.exit(msg)

        try:
            start = int(start)
        except ValueError as detail:
            msg = "Error reading line: {}. The start field is not " \
                  "an integer.\nError message: {}".format(line_number, detail)
            sys.exit(msg)

        try:
            end = int(end)
        except ValueError as detail:
            msg = "Error reading line: {}. The end field is not " \
                  "an integer.\nError message: {}".format(line_number, detail)
            sys.exit(msg)

        if prev_chrom == chrom:
            assert prev_start <= start, \
                "BED file not sorted. Please use a sorted bed file.\n{}{} ".format(prev_line, line)

        if chrom not in interval_tree:
            interval_tree[chrom] = IntervalTree()

        value = None

        if len(fields) > 3:
            value = fields[3:]
            try:
                line_min = min(map(float, value))
                if line_min < min_value:
                    min_value = line_min

                line_max = max(map(float, value))
                if line_max > max_value:
                    max_value = line_max
            except ValueError:
                pass

        assert end > start, "Start position larger or equal than end for line\n{} ".format(line)

        interval_tree[chrom].add(Interval(start, end, value))
        valid_intervals += 1

    if valid_intervals == 0:
        log.warning("No valid intervals were found in file {}".format(file_name))
    file_h.close()

    return interval_tree, min_value, max_value


class ReadBed(object):
    """
    Reads a bed file. Based on the number of fields
    it tries to guess the type of bed file used. Current options
    are bed3, bed6 and bed12

    Example:
    bed = readBed(open("file.bed", 'r'))
    for interval in bed:
        print(interval['start'])

    """

    def __init__(self, file_handle):

        self.file_type = None
        self.file_handle = file_handle
        self.line_number = 0
        # guess file type
        fields = self.get_no_comment_line()
        fields = to_string(fields)
        fields = fields.split()

        self.guess_file_type(fields)
        if type(file_handle) is types.GeneratorType:
            self._file_name = "<from generator>"
        else:
            self._file_name = file_handle.name
            self.file_handle.seek(0)
        self.prev_chrom = None
        self.prev_start = -1
        self.prev_line = None

        # list of bed fields
        self.fields = ['chromosome', 'start', 'end',
                       'name', 'score', 'strand',
                       'thick_start', 'thick_end',
                       'rgb', 'block_count',
                       'block_sizes', 'block_starts']

        if self.file_type == 'bed12':
            self.BedInterval = collections.namedtuple('BedInterval', self.fields)
        elif self.file_type == 'bed9':
            self.BedInterval = collections.namedtuple('BedInterval', self.fields[:9])
        else:
            self.BedInterval = collections.namedtuple('BedInterval', self.fields[:6])

    def __iter__(self):
        return self

    def get_no_comment_line(self):
        """
        Skips comment lines starting with '#'
        "track" or "browser" in the bed files
        """
        line = next(self.file_handle)
        line = to_string(line)
        if line.startswith("#") or line.startswith("track") or \
                line.startswith("browser") or line.strip() == '':
            line = self.get_no_comment_line()

        self.line_number += 1
        return line

    def guess_file_type(self, line_values):
        """try to guess type of bed file by counting the fields
        """
        if len(line_values) == 3:
            self.file_type = 'bed3'
        elif len(line_values) == 4:
            self.file_type = 'bedgraph'
        elif len(line_values) == 6:
            self.file_type = 'bed6'
        elif len(line_values) == 12:
            self.file_type = 'bed12'
        elif len(line_values) == 9:
            # this is a case where a specific color is encoded in the 10 field of the bed file
            self.file_type = 'bed9'
        elif len(line_values) > 6:
            # assume bed6
            self.file_type = 'bed6'
            log.warning("Number of fields in BED file is not standard. Assuming bed6\n")
        else:
            # assume bed3
            self.file_type = 'bed3'
            log.warning("Number of fields in BED file is not standard. Assuming bed3\n")
        return self.file_type

    def next(self):
        """
        Return
        ------
        bedInterval object
        """
        line = self.get_no_comment_line()

        bed = self.get_bed_interval(line)
        if self.prev_chrom == bed.chromosome:
            assert self.prev_start <= bed.start, \
                "BED file not sorted. Please use a sorted bed file.\n" \
                "File: {}\n" \
                "Previous line: {}\n Current line{} ".format(self._file_name, self.prev_line, line)

        self.prev_chrom = bed.chromosome
        self.prev_start = bed.start
        self.prev_line = line

        return bed

    def __next__(self):
        """
        Return
        ------
        bedInterval object
        """
        line = self.get_no_comment_line()

        bed = self.get_bed_interval(line)
        if self.prev_chrom == bed.chromosome:
            assert self.prev_start <= bed.start, \
                "BED file not sorted. Please use a sorted bed file.\n" \
                "File: {}\n" \
                "Previous line: {}\n Current line{} ".format(self._file_name, self.prev_line, line)

        self.prev_chrom = bed.chromosome
        self.prev_start = bed.start
        self.prev_line = line

        return bed

    def get_bed_interval(self, bed_line):
        r"""
        Processes each bed line from a bed file, casts the values and returns
        a namedtuple object

        >>> bed_line="chr1\t0\t1000\tgene_1\t0.5\t-\t0\t1000\t0\t3\t10,20,100\t20,200,700"
        >>> with open('/tmp/test.bed', 'w') as fh:
        ...     foo = fh.write(bed_line)
        >>> bed_f = ReadBed(open('/tmp/test.bed','r'))
        >>> bed = bed_f.get_bed_interval(bed_line)
        >>> bed.chromosome
        'chr1'
        >>> bed.block_starts
        [20, 200, 700]

        >>> bed_line="chr2\t0\t1000\tgene_1\t0.5\t-\n"
        >>> with open('/tmp/test.bed', 'w') as fh:
        ...     foo = fh.write(bed_line)
        >>> bed_f = ReadBed(open('/tmp/test.bed','r'))
        >>> bed_f.get_bed_interval(bed_line)
        BedInterval(chromosome='chr2', start=0, end=1000, name='gene_1', score=0.5, strand='-')
        """

        line_data = bed_line.strip()
        line_data = to_string(line_data)
        line_data = line_data.split("\t")

        if self.file_type == 'bed12':
            assert len(line_data) == 12, "File type detected is bed12 but line {}: {} does " \
                                         "not have 12 fields.".format(self.line_number, bed_line)

        elif self.file_type == 'bed3':
            assert len(line_data) == 3, "File type detected is bed3 but line {}: {} does " \
                                        "not have 3 fields.".format(self.line_number, bed_line)

        elif self.file_type == 'bed6':
            assert len(line_data) == 6, "File type detected is bed6 but line {}: {} does " \
                                        "not have 6 fields.".format(self.line_number, bed_line)
        line_values = []
        for idx, r in enumerate(line_data):
            # first field is always chromosome/contig name
            # and should be cast as a string
            # same for field 3 (name)
            if idx in [0, 3]:
                line_values.append(r)
            # check field strand
            elif idx == 5:
                if r not in ['+', '-', '.']:
                    if r == '1':
                        r = '+'
                    elif r == '-1':
                        r = '-'
                    else:
                        log.warning("*Warning, invalid strand value found {} for line #{}:\n{}\n "
                                    "Setting strand to '.'\n".format(r, bed_line, self.line_number))
                        r = '.'
                line_values.append(r)

            elif idx in [1, 2, 6, 7, 9]:
                # start and end fields must be integers, same for thichStart(6),
                # and thickEnd(7) and blockCount(9) fields
                try:
                    line_values.append(int(r))
                except ValueError:
                    log.warning("Value: {} in field {} at line {} is not an integer\n".format(r, idx + 1,
                                                                                              self.line_number))
                    return dict()
            # check item rgb
            elif idx == 8:
                r = to_string(r)
                rgb = r.split(",")
                if len(rgb) == 3:
                    try:
                        r = list(map(int, rgb))
                    except ValueError as detail:
                        log.warning("Error reading line: #{}. The rgb field {} is not "
                                    "valid.\nError message: {}\n".format(self.line_number, r, detail))
                line_values.append(r)

            elif idx in [10, 11]:
                # this are the block sizes and block start positions
                r = to_string(r)
                r_parts = r.split(',')
                try:
                    r = [int(x) for x in r_parts if x != '']
                except ValueError as detail:
                    log.warning("Error reading line #{}. The block field {} is not "
                                "valid.\nError message: {}\n".format(self.line_number, r, detail))
                line_values.append(r)

            else:
                try:
                    tmp = float(r)
                except ValueError:
                    tmp = r
                except TypeError:
                    tmp = r
                line_values.append(tmp)

        assert line_values[2] > line_values[1], \
            "Start position larger or equal than end for line #{}:\n{}\n".format(self.line_number,
                                                                                 bed_line)

        if self.file_type == 'bed3':
            line_values = line_values[0:3]
            # in case of a bed3, the id, score and strand
            # values are added as ".", 0, "." respectively
            line_values.extend([".", 0, "."])
        elif self.file_type == 'bed6':
            line_values = line_values[0:6]

        return self.BedInterval._make(line_values)


def bgz_bed(bed_path, bgz_path):
    cmd = ""
    if bed_path.endswith(".gz"):
        cmd += "zcat"
    else:
        cmd += "cat"
    subp.check_call(cmd + f" {bed_path} | sort -k1,1 -k2,2n | bgzip > {bgz_path}",
                    shell=True)
    return bgz_path


def index_bed(bgz_path):
    cmd = ['tabix', '-p', 'bed', bgz_path]
    subp.check_call(cmd)


def tabix_query(filename, chrom, start, end, split=True):
    """Call tabix and generate an array of strings for each line it returns."""
    query = '{}:{}-{}'.format(chrom, start, end)
    p = subp.Popen(['tabix', '-f', filename, query], stdout=subp.PIPE)
    for line in p.stdout:
        line = line.decode('utf-8')
        if not split:
            yield line
        else:
            yield line.strip().split('\t')


query_bed = partial(tabix_query, split=False)


def index_bedgraph(bgz_file):
    cmd = ['tabix', '-b', '2', '-e', '3', bgz_file]
    subp.check_call(cmd)


def build_bedgraph_bgz(file):
    file = osp.expanduser(file)
    if file.endswith(".bgz"):
        bgz_file = file
    else:
        bgz_file = file + '.bgz'
        log.info(f"Bgzip bedgraph file, save to {bgz_file}")
        bgz_bed(file, bgz_file)
    if not osp.exists(bgz_file + '.tbi'):
        log.info(f"Make tabix of bgz file, save to {bgz_file}.tbi")
        index_bedgraph(bgz_file)
    return bgz_file


def build_bed_index(file):
    file = osp.expanduser(file)
    if file.endswith(".bgz"):
        bgz_file = file
    else:
        bgz_file = file + '.bgz'
        log.info(f"Bgzip bed file, save to {bgz_file}")
        bgz_bed(file, bgz_file)
    if not osp.exists(bgz_file + '.tbi'):
        log.info(f"Make tabix of bgz file, save to {bgz_file}.tbi")
        index_bed(bgz_file)
    return bgz_file


def build_snp_index(file, col_chrom, col_pos):
    file = osp.expanduser(file)
    c = col_chrom + 1
    p = col_pos + 1
    if file.endswith(".bgz"):
        bgz_file = file
    elif osp.exists(file + '.bgz'):
        bgz_file = file + '.bgz'
    else:
        bgz_file = file + '.bgz'
        if file.endswith('.gz'):
            cmd = "zcat"
        else:
            cmd = "cat"
        cmd += f" {file} | sort -k{c},{c} -k{p},{p}n | bgzip > {bgz_file}"
        subp.check_call(cmd, shell=True)
    index_file = bgz_file + '.tbi'
    if not osp.exists(index_file):
        cmd = ['tabix', '-s', str(c), '-b', str(p), '-e', str(p), bgz_file]
        subp.check_call(cmd)
    return bgz_file


def bgz_bedpe(bedpe_path, bgz_path):
    if not osp.exists(bgz_path):
        cmd = f"sort -k1,1 -k4,4 -k2,2n -k5,5n {bedpe_path} | bgzip > {bgz_path}"
        subp.check_call(cmd, shell=True)


def index_bedpe(bgz_path):
    cmd = f"pairix -f -s 1 -d 4 -b 2 -e 3 -u 5 -v 6 {bgz_path}".split(" ")
    subp.check_call(cmd)


def pairix_query(bgz_file, query: GenomeRange, second: T.Optional[GenomeRange] = None,
                 open_region: bool = False, split: bool = True):
    if second:
        query = f"{query}|{second}"
    else:
        if open_region:
            query = f"{query}|{query.chrom}"
    cmd = ['pairix', str(bgz_file), str(query)]
    p = subp.Popen(cmd, stdout=subp.PIPE)
    for line in p.stdout:
        line = line.decode('utf-8')
        if not split:
            yield line
        else:
            yield line.strip().split('\t')


def process_bedpe(path):
    if path.endswith(".bgz"):
        bgz_file = path
    else:
        bgz_file = path + ".bgz"
        bgz_bedpe(path, bgz_file)
    if not osp.exists(f"{bgz_file}.px2"):
        index_bedpe(bgz_file)
    return bgz_file


def bgz_pairs(pairs_path, bgz_path):
    if not osp.exists(bgz_path):
        cmd = f"grep -v '#' {pairs_path} | sort -k2,2 -k4,4 -k3,3n -k5,5n | bgzip > {bgz_path}"
        subp.check_call(cmd, shell=True)


def index_pairs(bgz_path):
    cmd = f"pairix -f -p pairs {bgz_path}".split(" ")
    subp.check_call(cmd)


def process_pairs(path):
    if path.endswith(".bgz"):
        bgz_file = path
    else:
        bgz_file = path + ".bgz"
        bgz_pairs(path, bgz_file)
    if not osp.exists(f"{bgz_file}.px2"):
        index_pairs(bgz_file)
    return bgz_file


def process_gtf(gtf_path, out_path):
    cmd = f'(grep ^"#" {gtf_path}; grep -v ^"#" {gtf_path} | sort -k1,1 -k4,4n) | bgzip > {out_path}'
    subp.check_call(cmd, shell=True)


def gtf_gz_to_bgz(gz, bgz):
    cmd = f'gunzip -c {gz} | grep -v ^"#" | bgzip > {bgz}'
    subp.check_call(cmd, shell=True)


def tabix_index(filename, preset="gff"):
    """Call tabix to create an index for a bgzip-compressed file."""
    subp.check_call([
        'tabix', '-p', preset, filename
    ])


def build_gtf_index(file):
    file = osp.expanduser(file)
    if file.endswith(".gtf"):
        bgz_file = file + ".bgz"
        if not osp.exists(bgz_file):
            log.info(f"Process the gtf and do bgzip, save to {bgz_file}.")
            process_gtf(file, bgz_file)
    elif file.endswith(".gtf.gz"):
        bgz_file = file.rstrip(".gz") + ".bgz"
        log.info(f"Convert .gtf.gz to .gtf.bgz, save to {bgz_file}.")
        if not osp.exists(bgz_file):
            gtf_gz_to_bgz(file, bgz_file)
    elif file.endswith(".gtf.bgz"):
        bgz_file = file
    else:
        raise IOError(f"GTF track only support GTF file(.gtf or .gtf.gz), got {file}.")

    idx_file = bgz_file + ".tbi"
    if not osp.exists(idx_file):
        log.info(f"Tabix index not found, build it in {idx_file}")
        tabix_index(bgz_file)
    return bgz_file

